################################################################################
# Copyright (c) 2015-2019, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for the asynchronous Telescope State client."""

import asyncio
from unittest import mock

import asynctest
import async_timeout
import numpy as np
import fakeredis.aioredis

import katsdptelstate
from katsdptelstate import ImmutableKeyError, encode_value, KeyType, ENCODING_MSGPACK
from katsdptelstate.aio import TelescopeState
from katsdptelstate.aio.memory import MemoryBackend
from katsdptelstate.aio.redis import RedisBackend


class TestTelescopeState(asynctest.TestCase):
    async def setUp(self) -> None:
        self.ts = await self.make_telescope_state()
        self.ns = self.ts.view('ns')

    async def tearDown(self) -> None:
        self.ts.backend.close()
        await self.ts.backend.wait_closed()

    async def make_telescope_state(self) -> TelescopeState:
        return TelescopeState()

    def test_bad_construct(self) -> None:
        with self.assertRaises(ValueError):
            TelescopeState(self.ts.backend, base=self.ts)

    def test_namespace(self) -> None:
        self.assertEqual(self.ts.prefixes, ('',))
        self.assertEqual(self.ns.prefixes, ('ns_', ''))
        ns2 = self.ns.view(b'ns_child_grandchild')
        self.assertEqual(ns2.prefixes, ('ns_child_grandchild_', 'ns_', ''))
        self.assertEqual(ns2.root().prefixes, ('',))
        ns_excl = self.ns.view('exclusive', exclusive=True)
        self.assertEqual(ns_excl.prefixes, ('exclusive_',))

    async def test_basic_add(self) -> None:
        await self.ts.add('test_key', 1234.5)
        self.assertEqual(await self.ts['test_key'], 1234.5)

        await self.ts.delete('test_key')
        with self.assertRaises(KeyError):
            await self.ts['test_key']

    async def test_namespace_add(self) -> None:
        await self.ns.add('test_key', 1234.5)
        self.assertEqual(await self.ns['test_key'], 1234.5)
        self.assertEqual(await self.ts[self.ts.join('ns', 'test_key')], 1234.5)
        with self.assertRaises(KeyError):
            await self.ts['test_key']

    async def test_delete(self) -> None:
        await self.ts.add('test_key', 1234.5)
        self.assertTrue(await self.ts.exists('test_key'))
        await self.ts.delete('test_key')
        await self.ts.delete('test_key')
        self.assertFalse(await self.ts.exists('test_key'))

    async def test_namespace_delete(self) -> None:
        await self.ts.add('parent_key', 1234.5)
        await self.ns.add('child_key', 2345.6)
        await self.ns.delete('child_key')
        await self.ns.delete('parent_key')
        self.assertEqual(await self.ts.keys(), [])

    async def test_clear(self) -> None:
        await self.ts.add('test_key', 1234.5)
        await self.ts.add('test_key_rt', 2345.6)
        await self.ts.clear()
        self.assertEqual(await self.ts.keys(), [])

    async def test_get_default(self) -> None:
        self.assertIsNone(await self.ts.get('foo'))
        self.assertEqual(await self.ts.get('foo', 'bar'), 'bar')

    async def test_get_return_encoded(self) -> None:
        for immutable in [True, False]:
            x = np.array([(1.0, 2), (3.0, 4)], dtype=[('x', float), ('y', int)])
            await self.ts.add('test_key_rt', x, immutable=True)
            x_decoded = await self.ts.get('test_key_rt')
            self.assertTrue((x_decoded == x).all())
            x_encoded = await self.ts.get('test_key_rt', return_encoded=True)
            self.assertEqual(x_encoded, encode_value(x))
            await self.ts.delete('test_key_rt')

    async def test_get_range_return_encoded(self) -> None:
        test_values = ['Test Value: {}'.format(x) for x in range(5)]
        for i, test_value in enumerate(test_values):
            await self.ts.add('test_key', test_value, i)
        stored_values = await self.ts.get_range('test_key', st=0)
        self.assertEqual(stored_values[2][0], test_values[2])
        stored_values_pickled = await self.ts.get_range('test_key', st=0, return_encoded=True)
        self.assertEqual(stored_values_pickled[2][0], encode_value(test_values[2]))
        # check timestamp
        self.assertEqual(stored_values_pickled[2][1], 2)

    async def test_immutable(self) -> None:
        await self.ts.add('test_immutable', 1234.5, immutable=True)
        with self.assertRaises(ImmutableKeyError):
            await self.ts.add('test_immutable', 1234.5)
        with self.assertRaises(ImmutableKeyError):
            await self.ts.add('test_immutable', 5432.1, immutable=True)
        with self.assertRaises(ImmutableKeyError):
            await self.ts.set_indexed('test_immutable', 1234.5, 1234.5)

    async def test_immutable_same_value(self) -> None:
        await self.ts.add('test_immutable', 1234.5, immutable=True)
        await self.ts.add('test_mutable', 1234.5)
        with self.assertRaises(ImmutableKeyError):
            await self.ts.add('test_mutable', 2345.6, immutable=True)

    async def test_immutable_same_value_str(self) -> None:
        with mock.patch('katsdptelstate.encoding._allow_pickle', True), \
                mock.patch('katsdptelstate.encoding._warn_on_pickle', False):
            await self.ts.add('test_bytes', b'caf\xc3\xa9', immutable=True)
            await self.ts.add('test_bytes', b'caf\xc3\xa9', immutable=True)
            await self.ts.add('test_bytes', 'café', immutable=True)
            with self.assertRaises(ImmutableKeyError):
                await self.ts.add('test_bytes', b'cafe', immutable=True)
            with self.assertRaises(ImmutableKeyError):
                await self.ts.add('test_bytes', 'cafe', immutable=True)
            await self.ts.add('test_unicode', 'ümlaut', immutable=True)
            await self.ts.add('test_unicode', 'ümlaut', immutable=True)
            await self.ts.add('test_unicode', b'\xc3\xbcmlaut', immutable=True)
            with self.assertRaises(ImmutableKeyError):
                await self.ts.add('test_unicode', b'umlaut', immutable=True)
            with self.assertRaises(ImmutableKeyError):
                await self.ts.add('test_unicode', 'umlaut', immutable=True)
            # Test with a binary string that isn't valid UTF-8
            await self.ts.add('test_binary', b'\x00\xff', immutable=True)
            await self.ts.add('test_binary', b'\x00\xff', immutable=True)
            # Test Python 2/3 interop by directly injecting the pickled values
            await self.ts.backend.set_immutable(b'test_2', b"S'hello'\np1\n.")
            await self.ts.backend.set_immutable(b'test_3', b'Vhello\np0\n.')
            await self.ts.add('test_2', 'hello', immutable=True)
            await self.ts.add('test_3', 'hello', immutable=True)
            # Test handling of the case where the old value cannot be decoded
            # Empty string is never valid encoding
            await self.ts.backend.set_immutable(b'test_failed_decode', b'')
            with self.assertRaisesRegex(ImmutableKeyError, 'failed to decode the previous value'):
                await self.ts.add('test_failed_decode', '', immutable=True)

    async def test_immutable_none(self) -> None:
        await self.ts.add('test_none', None, immutable=True)
        self.assertIsNone(await self.ts.get('test_none'))
        self.assertIsNone(await self.ts.get('test_none', 'not_none'))
        self.assertIsNone(await self.ts['test_none'])

    async def test_immutable_wrong_type(self) -> None:
        await self.ts.add('test_mutable', 5)
        await self.ts.add('test_indexed', 5, 5)
        with self.assertRaises(ImmutableKeyError):
            await self.ts.add('test_mutable', 5, immutable=True)
        with self.assertRaises(ImmutableKeyError):
            await self.ts.add('test_indexed', 5, immutable=True)

    async def test_namespace_immutable(self) -> None:
        await self.ts.add('parent_immutable', 1234.5, immutable=True)
        await self.ns.add('child_immutable', 2345.5, immutable=True)
        with self.assertRaises(KeyError):
            await self.ts['child_immutable']
        self.assertEqual(await self.ns.get('child_immutable'), 2345.5)
        self.assertEqual(await self.ns.get('parent_immutable'), 1234.5)

    async def test_key_type(self) -> None:
        await self.ts.add('parent_immutable', 1, immutable=True)
        await self.ns.add('child_immutable', 2, immutable=True)
        await self.ts.add('parent', 3)
        await self.ns.add('child', 4)
        await self.ts.set_indexed('parent_indexed', 'a', 1)
        await self.ns.set_indexed('child_indexed', 'b', 2)

        self.assertEqual(await self.ts.key_type('parent_immutable'), KeyType.IMMUTABLE)
        self.assertEqual(await self.ts.key_type('parent'), KeyType.MUTABLE)
        self.assertEqual(await self.ts.key_type('parent_indexed'), KeyType.INDEXED)
        self.assertEqual(await self.ts.key_type('child_immutable'), None)
        self.assertEqual(await self.ts.key_type('child'), None)
        self.assertEqual(await self.ts.key_type('not_a_key'), None)

        self.assertEqual(await self.ns.key_type('parent_immutable'), KeyType.IMMUTABLE)
        self.assertEqual(await self.ns.key_type('parent'), KeyType.MUTABLE)
        self.assertEqual(await self.ns.key_type('parent_indexed'), KeyType.INDEXED)
        self.assertEqual(await self.ns.key_type('child_immutable'), KeyType.IMMUTABLE)
        self.assertEqual(await self.ns.key_type('child'), KeyType.MUTABLE)
        self.assertEqual(await self.ns.key_type('child_indexed'), KeyType.INDEXED)
        self.assertEqual(await self.ns.key_type('not_a_key'), None)

    async def test_keys(self) -> None:
        await self.ts.add('key1', 'a')
        await self.ns.add('key2', 'b')
        await self.ns.add(b'key2', 'c')
        await self.ts.add(b'immutable', 'd', immutable=True)
        self.assertEqual(await self.ts.keys(), ['immutable', 'key1', 'ns_key2'])
        self.assertEqual(await self.ts.keys('ns_*'), ['ns_key2'])

    async def test_complex_store(self) -> None:
        x = np.array([(1.0, 2), (3.0, 4)], dtype=[('x', float), ('y', int)])
        await self.ts.add('test_key', x)
        self.assertTrue((await self.ts['test_key'] == x).all())

    async def test_exists(self) -> None:
        await self.ts.add('test_key', 1234.5)
        self.assertTrue(await self.ts.exists('test_key'))
        self.assertFalse(await self.ts.exists('nonexistent_test_key'))

    async def test_time_range(self) -> None:
        await self.ts.delete('test_key')
        await self.ts.add('test_key', 8192, 1)
        await self.ts.add('test_key', 16384, 2)
        await self.ts.add('test_key', 4096, 3)
        await self.ts.add('test_key', 2048, 4)
        await self.ts.add('test_immutable', 12345, immutable=True)
        self.assertEqual([(2048, 4)], await self.ts.get_range('test_key'))
        self.assertEqual([(16384, 2)], await self.ts.get_range('test_key', et=3))
        self.assertEqual([(8192, 1), (16384, 2), (4096, 3)],
                         await self.ts.get_range('test_key', st=2, et=4, include_previous=True))
        self.assertEqual([(8192, 1), (16384, 2), (4096, 3), (2048, 4)],
                         await self.ts.get_range('test_key', st=0))
        self.assertEqual([(8192, 1), (16384, 2), (4096, 3)],
                         await self.ts.get_range('test_key', st=0, et=3.5))
        self.assertEqual([(8192, 1)], await self.ts.get_range('test_key', st=-1, et=1.5))
        self.assertEqual([(16384, 2), (4096, 3), (2048, 4)],
                         await self.ts.get_range('test_key', st=2))
        self.assertEqual([(8192, 1)], await self.ts.get_range('test_key', et=1.5))
        self.assertEqual([], await self.ts.get_range('test_key', 3.5, 1.5))
        self.assertEqual([], await self.ts.get_range('test_key', et=-0.))
        self.assertEqual([(8192, 1), (16384, 2), (4096, 3), (2048, 4)],
                         await self.ts.get_range('test_key', st=1.5, include_previous=True))
        self.assertEqual([(2048, 4)],
                         await self.ts.get_range('test_key', st=5, et=6, include_previous=True))
        self.assertEqual([(8192, 1), (16384, 2), (4096, 3)],
                         await self.ts.get_range('test_key', st=2, et=4, include_previous=True))
        with self.assertRaises(KeyError):
            await self.ts.get_range('not_a_key')
        with self.assertRaises(ImmutableKeyError):
            await self.ts.get_range('test_immutable')

    async def test_time_range_include_end(self) -> None:
        await self.ts.add('test_key', 1234, 0)
        await self.ts.add('test_key', 8192, 1)
        await self.ts.add('test_key', 16384, 2)
        await self.ts.add('test_key', 4096, 3)
        await self.ts.add('test_key', 2048, 4)
        self.assertEqual([], await self.ts.get_range('test_key', st=1.5, et=1.5, include_end=True))
        self.assertEqual([(1234, 0)],
                         await self.ts.get_range('test_key', st=0.0, et=0.0, include_end=True))
        self.assertEqual([(1234, 0)],
                         await self.ts.get_range('test_key', st=0.0, et=-0.0, include_end=True))
        self.assertEqual([(4096, 3), (2048, 4)],
                         await self.ts.get_range('test_key', st=3, et=4, include_end=True))
        # include_previous tests
        self.assertEqual([(8192, 1), (16384, 2)],
                         await self.ts.get_range('test_key', et=2, include_end=True))
        self.assertEqual([(8192, 1), (16384, 2), (4096, 3)],
                         await self.ts.get_range('test_key', st=2, et=3,
                                                 include_previous=True, include_end=True))

    async def test_add_duplicate(self) -> None:
        await self.ts.add('test_key', 'value', 1234.5)
        await self.ts.add('test_key', 'value', 1234.5)
        self.assertEqual([('value', 1234.5)], await self.ts.get_range('test_key', st=0))

    async def test_wait_key_already_done_mutable(self) -> None:
        """Calling wait_key with a condition that is met must return (mutable version)."""
        await self.ts.add('test_key', 123)
        value, timestamp = (await self.ts.get_range('test_key'))[0]
        await self.ts.wait_key('test_key', lambda v, t: v == value and t == timestamp)

    async def test_wait_key_already_done_immutable(self) -> None:
        """Calling wait_key with a condition that is met must return (immutable version)."""
        await self.ts.add('test_key', 123, immutable=True)
        await self.ts.wait_key('test_key', lambda v, t: v == 123 and t is None)

    async def test_wait_key_already_done_indexed(self) -> None:
        """Calling wait_key with a condition that is met must return (indexed version)."""
        await self.ts.set_indexed('test_key', 'idx', 5)
        await self.ts.wait_key('test_key', lambda v, t: v == {'idx': 5} and t is None)

    async def test_wait_key_timeout(self) -> None:
        """wait_key must time out in the given time if the condition is not met"""
        with self.assertRaises(asyncio.TimeoutError):
            with async_timeout.timeout(0.1):
                await self.ts.wait_key('test_key')
        with self.assertRaises(asyncio.TimeoutError):
            with async_timeout.timeout(0.1):
                # Takes a different code path, even though equivalent
                await self.ts.wait_key('test_key', lambda value, ts: True)

    async def _set_key_immutable(self) -> None:
        await asyncio.sleep(0.1)
        await self.ts.set('test_key', 123)

    async def _set_key_mutable(self) -> None:
        await asyncio.sleep(0.1)
        await self.ts.add('test_key', 123, ts=1234567890)

    async def _set_key_indexed(self) -> None:
        await asyncio.sleep(0.1)
        await self.ts.set_indexed('test_key', 'idx', 123)

    async def test_wait_key_delayed(self) -> None:
        """wait_key must succeed with a timeout that does not expire before the condition is met."""
        for (set_key, value, timestamp) in [
                (self._set_key_mutable, 123, 1234567890),
                (self._set_key_immutable, 123, None),
                (self._set_key_indexed, {'idx': 123}, None)]:
            task = asyncio.ensure_future(set_key())
            await self.ts.wait_key('test_key', lambda v, t: v == value and t == timestamp)
            self.assertEqual(value, await self.ts.get('test_key'))
            await task
            await self.ts.delete('test_key')

    async def test_wait_key_delayed_unconditional(self) -> None:
        """wait_key must succeed when given a timeout that does not expire before key appears."""
        for set_key, value in [
                (self._set_key_mutable, 123),
                (self._set_key_immutable, 123),
                (self._set_key_indexed, {'idx': 123})]:
            task = asyncio.ensure_future(set_key())
            await self.ts.wait_key('test_key')
            self.assertEqual(value, await self.ts['test_key'])
            await task
            await self.ts.delete('test_key')

    async def test_wait_key_cancel(self) -> None:
        """wait_key must deal gracefully with cancellation."""
        task = asyncio.ensure_future(self.ts.wait_key('test_key'))
        await asyncio.sleep(0.1)
        task.cancel()
        with self.assertRaises(asyncio.CancelledError):
            await task

    async def test_wait_key_shadow(self) -> None:
        """updates to a shadowed qualified key must be ignored"""
        async def set_key(telstate):
            await asyncio.sleep(0.1)
            await telstate.add('test_key', True, immutable=True)

        ns2 = self.ns.view('ns2')
        # Put a non-matching key into a mid-level namespace
        await self.ns.add('test_key', False)
        # Put a matching key into a shadowed namespace after a delay
        task = asyncio.ensure_future(set_key(self.ts))
        with self.assertRaises(asyncio.TimeoutError):
            with async_timeout.timeout(0.5):
                await ns2.wait_key('test_key', lambda value, ts: value is True)
        await task

        # Put a matching key into a non-shadowed namespace after a delay
        task = asyncio.ensure_future(set_key(ns2))
        with async_timeout.timeout(5):
            await ns2.wait_key('test_key', lambda value, ts: value is True)
        await task

    async def test_wait_indexed_already_done(self) -> None:
        await self.ts.set_indexed('test_key', 'sub_key', 5)
        await self.ts.wait_indexed('test_key', 'sub_key')
        await self.ts.wait_indexed('test_key', 'sub_key', lambda v: v == 5)

    async def test_wait_indexed_timeout(self) -> None:
        await self.ts.set_indexed('test_key', 'sub_key', 5)
        with self.assertRaises(asyncio.TimeoutError):
            with async_timeout.timeout(0.05):
                await self.ts.wait_indexed('test_key', 'sub_key', lambda v: v == 4)
        with self.assertRaises(asyncio.TimeoutError):
            with async_timeout.timeout(0.05):
                await self.ts.wait_indexed('test_key', 'another_key')
        with self.assertRaises(asyncio.TimeoutError):
            with async_timeout.timeout(0.05):
                await self.ts.wait_indexed('not_present', 'sub_key')

    async def test_wait_indexed_delayed(self) -> None:
        async def set_key():
            await self.ts.set_indexed('test_key', 'foo', 1)
            await asyncio.sleep(0.05)
            await self.ts.set_indexed('test_key', 'bar', 2)
        for condition in [None, lambda value: value == 2]:
            await self.ts.delete('test_key')
            task = asyncio.ensure_future(set_key())
            with async_timeout.timeout(2):
                await self.ts.wait_indexed('test_key', 'bar', condition)
            self.assertEqual(2, await self.ts.get_indexed('test_key', 'bar'))
            await task

    async def test_wait_indexed_cancel(self) -> None:
        task = asyncio.ensure_future(self.ts.wait_indexed('test_key', 'sub_key'))
        await asyncio.sleep(0.1)
        task.cancel()
        with self.assertRaises(asyncio.CancelledError):
            await task

    async def test_wait_indexed_shadow(self) -> None:
        async def set_key(telstate):
            await asyncio.sleep(0.1)
            await telstate.set_indexed('test_key', 'sub_key', 1)

        # Shadow the key that will be set by set_key
        await self.ns.set_indexed('test_key', 'blah', 1)
        task = asyncio.ensure_future(set_key(self.ts))
        with self.assertRaises(asyncio.TimeoutError):
            with async_timeout.timeout(0.2):
                await self.ns.wait_indexed('test_key', 'sub_key')
        await task

    async def test_wait_indexed_wrong_type(self) -> None:
        async def set_key():
            await asyncio.sleep(0.1)
            await self.ts.add('test_key', 'value')

        task = asyncio.ensure_future(set_key())
        with self.assertRaises(ImmutableKeyError):
            await self.ts.wait_indexed('test_key', 'value')
        await task
        # Different code-path when key was already set at the start
        with self.assertRaises(ImmutableKeyError):
            await self.ts.wait_indexed('test_key', 'value')

    async def _test_mixed_unicode_bytes(self, ns, key) -> None:
        await self.ts.clear()
        await ns.add(key, 'value', immutable=True)
        self.assertEqual(await ns.get(key), 'value')
        self.assertTrue(await ns.exists(key))
        self.assertEqual(await ns.key_type(key), KeyType.IMMUTABLE)
        await ns.delete(key)
        await ns.add(key, 'value1', ts=1)
        self.assertEqual(await ns.get_range(key), [('value1', 1.0)])
        await ns.wait_key(key)

    async def test_mixed_unicode_bytes(self) -> None:
        await self._test_mixed_unicode_bytes(self.ts.view(b'ns'), 'test_key')
        await self._test_mixed_unicode_bytes(self.ts.view('ns'), b'test_key')

    async def test_undecodable_bytes_in_key(self) -> None:
        """Gracefully handle non-UTF-8 bytes in keys."""
        key_b = b'undecodable\xff'
        await self.ts.backend.set_immutable(key_b, encode_value('hello'))
        key = [k for k in await self.ts.keys() if k.startswith('undecodable')][0]
        self.assertEqual(await self.ts.get(key), 'hello')
        self.assertEqual(await self.ts.get(key_b), 'hello')
        self.assertEqual(await self.ts.get(key_b[1:]), None)

    async def test_get_indexed(self) -> None:
        await self.ts.set_indexed('test_indexed', 'a', 1)
        await self.ts.set_indexed('test_indexed', (2, 3j), 2)
        self.assertEqual(await self.ts.get_indexed('test_indexed', 'a'), 1)
        self.assertEqual(await self.ts.get_indexed('test_indexed', (2, 3j)), 2)
        self.assertIsNone(await self.ts.get_indexed('test_indexed', 'missing'))
        self.assertIsNone(await self.ts.get_indexed('not_a_key', 'missing'))
        self.assertEqual(await self.ts.get_indexed('test_indexed', 'missing', 'default'), 'default')
        self.assertEqual(await self.ts.get_indexed('not_a_key', 'missing', 'default'), 'default')
        self.assertEqual(await self.ts.get('test_indexed'), {'a': 1, (2, 3j): 2})

    async def test_indexed_return_encoded(self) -> None:
        await self.ts.set_indexed('test_indexed', 'a', 1)
        values = await self.ts.get('test_indexed', return_encoded=True)
        self.assertEqual(values, {
            encode_value('a', encoding=ENCODING_MSGPACK): encode_value(1)
        })
        self.assertEqual(await self.ts.get_indexed('test_indexed', 'a', return_encoded=True),
                         encode_value(1))

    async def test_set_indexed_immutable(self) -> None:
        await self.ts.set_indexed('test_indexed', 'a', 1)
        await self.ts.set_indexed('test_indexed', 'a', 1)  # Same value is okay
        with self.assertRaises(ImmutableKeyError):
            await self.ts.set_indexed('test_indexed', 'a', 2)
        self.assertEqual(await self.ts.get_indexed('test_indexed', 'a'), 1)

    async def test_set_indexed_unhashable(self) -> None:
        with self.assertRaises(TypeError):
            await self.ts.set_indexed('test_indexed', ['list', 'is', 'not', 'hashable'], 0)

    async def test_indexed_wrong_type(self) -> None:
        await self.ts.set('test_immutable', 1)
        await self.ts.add('test_mutable', 2)
        with self.assertRaises(ImmutableKeyError):
            await self.ts.set_indexed('test_immutable', 'a', 1)
        with self.assertRaises(ImmutableKeyError):
            await self.ts.set_indexed('test_mutable', 'a', 1)
        with self.assertRaises(ImmutableKeyError):
            await self.ts.get_indexed('test_immutable', 'a')
        with self.assertRaises(ImmutableKeyError):
            await self.ts.get_indexed('test_mutable', 'a')

    async def test_namespace_indexed(self) -> None:
        await self.ts.set_indexed('test_indexed', 'a', 1)
        self.assertEqual(await self.ns.get('test_indexed'), {'a': 1})
        self.assertEqual(await self.ns.get_indexed('test_indexed', 'a'), 1)
        await self.ns.set_indexed('test_indexed', 'b', 2)
        self.assertEqual(await self.ns.get('test_indexed'), {'b': 2})
        self.assertEqual(await self.ns.get_indexed('test_indexed', 'b'), 2)
        # Namespace key must completely shadow root
        self.assertIsNone(await self.ns.get_indexed('test_indexed', 'a'))


class TestTelescopeStateRedis(TestTelescopeState):
    async def make_telescope_state(self) -> TelescopeState:
        client = await fakeredis.aioredis.create_redis_pool()
        return TelescopeState(RedisBackend(client))


class TestSharedMemoryBackend(asynctest.TestCase):
    def setUp(self) -> None:
        self.async_backend = MemoryBackend()
        self.sync_backend = self.async_backend.to_sync()
        self.async_ts = TelescopeState(self.async_backend)
        self.sync_ts = katsdptelstate.TelescopeState(self.sync_backend)

    async def test_shared(self) -> None:
        self.sync_ts['foo'] = 'bar'
        self.assertEqual(await self.async_ts.get('foo'), 'bar')

    async def test_from_sync(self) -> None:
        ts = TelescopeState(MemoryBackend.from_sync(self.sync_backend))
        self.sync_ts['foo'] = 'bar'
        self.assertEqual(await ts.get('foo'), 'bar')



################################################################################
# Copyright (c) 2015-2019, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import socket
import struct
from typing import List, Iterable, Iterator, Callable, Any

import ipaddress
import netifaces


class Endpoint:
    """A TCP or UDP endpoint consisting of a host and a port.

    Typically the host should be a string (whether a hostname or IP address) and
    the port should be an integer, but users are free to use other conventions.
    """

    def __init__(self, host: Any, port: Any) -> None:
        self.host = host
        self.port = port

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Endpoint) and self.host == other.host and self.port == other.port

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __hash__(self) -> int:
        return hash((self.host, self.port))

    def __str__(self) -> str:
        if ':' in self.host:
            # IPv6 address - escape it
            return '[{}]:{}'.format(self.host, self.port)
        else:
            return '{}:{}'.format(self.host, self.port)

    def __repr__(self) -> str:
        return 'Endpoint({!r}, {!r})'.format(self.host, self.port)

    def __iter__(self) -> Iterator:
        """Support `tuple(endpoint)` for passing to a socket function"""
        return iter((self.host, self.port))

    def multicast_subscribe(self, sock: socket.socket) -> bool:
        """If the address is an IPv4 multicast address, subscribe to the group on `sock`.

        Return `True` if the host is a multicast address.
        """
        try:
            raw = socket.inet_aton(str(self.host))
        except OSError:
            return False
        else:
            # IPv4 multicast is the range 224.0.0.0 - 239.255.255.255
            if raw[0] >= 224 and raw[0] <= 239:
                for iface in netifaces.interfaces():
                    for addr in netifaces.ifaddresses(iface).get(netifaces.AF_INET, []):
                        # Skip point-to-point links (includes loopback)
                        if 'peer' in addr:
                            continue
                        if_raw = socket.inet_aton(addr['addr'])
                        mreq = struct.pack("4s4s", raw, if_raw)
                        sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)
                        break  # Should only need to subscribe once per interface
                return True
            else:
                return False


def endpoint_parser(default_port: Any) -> Callable[[str], Endpoint]:
    """Return a factory function that parses a string.

    The string is either `hostname`, or `hostname:port`, where `port` is an
    integer. IPv6 addresses are written in square brackets (similar to RFC
    2732) to disambiguate the embedded colons.
    """
    def parser(text: str) -> Endpoint:
        port = default_port
        # Find the last :, which should separate the port
        pos = text.rfind(':')
        # If the host starts with a bracket, do not match a : inside the
        # brackets.
        if len(text) and text[0] == '[':
            right = text.find(']')
            if right != -1:
                pos = text.rfind(':', right + 1)
        if pos != -1:
            host = text[:pos]
            port = int(text[pos + 1:])
        else:
            host = text
        # Strip the []
        if len(host) and host[0] == '[' and host[-1] == ']':
            # Validate the IPv6 address
            host = host[1:-1]
            try:
                socket.inet_pton(socket.AF_INET6, host)
            except OSError as e:
                raise ValueError(str(e))
        return Endpoint(host, port)
    return parser


def endpoint_list_parser(default_port: Any,
                         single_port: bool = False) -> Callable[[str], List[Endpoint]]:
    """Return a factory function that parses a string.

    The string comprises a comma-separated list, each element of which is of
    the form taken by :func:`endpoint_parser`. Optionally, the hostname may be
    followed by `+count`, where `count` is an integer specifying a number of
    sequential IP addresses (in addition to the explicitly named one). This
    variation is only valid with IPv4 addresses.

    If `single_port` is true, then it will reject any list that contains
    more than one distinct port number, as well as an empty list. This allows
    the user to determine a unique port for the list.
    """
    def parser(text: str) -> List[Endpoint]:
        sub_parser = endpoint_parser(default_port)
        parts = text.split(',')
        endpoints = []
        for part in parts:
            endpoint = sub_parser(part.strip())
            pos = endpoint.host.rfind('+')
            if pos != -1:
                start = endpoint.host[:pos]
                count = int(endpoint.host[pos + 1:])
                if count < 0:
                    raise ValueError('bad count {}'.format(count))
                try:
                    start_raw = struct.unpack('>I', socket.inet_aton(start))[0]
                    for i in range(start_raw, start_raw + count + 1):
                        host = socket.inet_ntoa(struct.pack('>I', i))
                        endpoints.append(Endpoint(host, endpoint.port))
                except OSError:
                    raise ValueError('invalid IPv4 address in {}'.format(start))
            else:
                endpoints.append(endpoint)
        if single_port:
            if not endpoints:
                raise ValueError('empty list')
            else:
                for endpoint in endpoints:
                    if endpoint.port != endpoints[0].port:
                        raise ValueError('all endpoints must use the same port')
        return endpoints
    return parser


def endpoints_to_str(endpoints: Iterable[Endpoint]) -> str:
    """Convert a list of endpoints into a compact string that generates the same list.

    This is the inverse of :func:`katsdptelstate.endpoint.endpoint_list_parser`.
    """
    # Partition the endpoints by type
    ipv4 = []
    ipv6 = []
    other = []
    for endpoint in endpoints:
        # ipaddress module requires unicode, so convert if not already
        host = endpoint.host.decode('utf-8') if isinstance(endpoint.host, bytes) else endpoint.host
        try:
            ipv4.append(Endpoint(ipaddress.IPv4Address(host), endpoint.port))
        except ipaddress.AddressValueError:
            try:
                ipv6.append(Endpoint(ipaddress.IPv6Address(host), endpoint.port))
            except ipaddress.AddressValueError:
                other.append(endpoint)
    # We build a list of parts, each of which is either host:port, addr:port or
    # addr+n:port (where :port is omitted if None). These get comma-separated
    # at the end.
    parts = []
    for ip in (ipv4, ipv6):
        # lists of address, num, port (not tuples because mutated)
        ip_parts = []      # type: List[List[Any]]
        # Group endpoints with the same port together, then order by IP address
        ip.sort(key=lambda endpoint: (endpoint.port is not None, endpoint.port, endpoint.host))
        for endpoint in ip:
            if (ip_parts and ip_parts[-1][2] == endpoint.port
                    and ip_parts[-1][0] + ip_parts[-1][1] == endpoint.host):
                ip_parts[-1][1] += 1
            else:
                ip_parts.append([endpoint.host, 1, endpoint.port])
        for (address, num, port) in ip_parts:
            if ip is ipv6:
                s = '[' + address.compressed + ']'
            else:
                s = address.compressed
            if num > 1:
                s += '+{}'.format(num - 1)
            if port is not None:
                s += ':{}'.format(port)
            parts.append(s)
    for endpoint in other:
        s = str(endpoint.host)
        if endpoint.port is not None:
            s += ':{}'.format(endpoint.port)
        parts.append(s)
    return ','.join(parts)



################################################################################
# Copyright (c) 2015-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

from typing import Optional

from .utils import _PathType


class TelstateError(RuntimeError):
    """Base class for errors from this package."""


class ConnectionError(TelstateError):
    """The initial connection to the Redis server failed."""


class RdbParseError(TelstateError):
    """Error parsing RDB file."""

    def __init__(self, filename: Optional[_PathType] = None) -> None:
        self.filename = filename

    def __str__(self) -> str:
        name = repr(self.filename) if self.filename else 'object'
        return 'Invalid RDB file {}'.format(name)


class InvalidKeyError(TelstateError):
    """A key collides with a class attribute.

    This is kept only for backwards compatibility. It is no longer considered an error.
    """


class InvalidTimestampError(TelstateError):
    """Negative or non-finite timestamp"""


class ImmutableKeyError(TelstateError):
    """An attempt was made to modify an immutable key"""


class TimeoutError(TelstateError):
    """A wait for a key timed out"""


class CancelledError(TelstateError):
    """A wait for a key was cancelled"""


class DecodeError(ValueError, TelstateError):
    """An encoded value found in telstate could not be decoded"""


class EncodeError(ValueError, TelstateError):
    """A value could not be encoded"""



################################################################################
# Copyright (c) 2015-2019, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for the sdp telescope state client."""

import threading
import time
import unittest
from unittest import mock

import numpy as np
import fakeredis

from katsdptelstate import (TelescopeState, ImmutableKeyError,
                            TimeoutError, CancelledError, encode_value, KeyType,
                            ENCODING_MSGPACK)
from katsdptelstate.memory import MemoryBackend


class TestTelescopeState(unittest.TestCase):
    def setUp(self) -> None:
        self.ts = self.make_telescope_state()
        self.ns = self.ts.view('ns')

    def make_telescope_state(self) -> TelescopeState:
        return TelescopeState()

    def test_bad_construct(self) -> None:
        with self.assertRaises(ValueError):
            TelescopeState('redis.example.com:7148', base=self.ts)
        with self.assertRaises(ValueError):
            TelescopeState('', 1, base=self.ts)
        with self.assertRaises(ValueError):
            TelescopeState(MemoryBackend(), 1)
        with self.assertRaises(ValueError):
            TelescopeState(db=1)

    def test_namespace(self) -> None:
        self.assertEqual(self.ts.prefixes, ('',))
        self.assertEqual(self.ns.prefixes, ('ns_', ''))
        ns2 = self.ns.view(b'ns_child_grandchild')
        self.assertEqual(ns2.prefixes, ('ns_child_grandchild_', 'ns_', ''))
        self.assertEqual(ns2.root().prefixes, ('',))
        ns_excl = self.ns.view('exclusive', exclusive=True)
        self.assertEqual(ns_excl.prefixes, ('exclusive_',))

    def test_basic_add(self) -> None:
        self.ts.add('test_key', 1234.5)
        self.assertEqual(self.ts.test_key, 1234.5)
        self.assertEqual(self.ts['test_key'], 1234.5)

        self.ts.delete('test_key')
        with self.assertRaises(AttributeError):
            self.ts.test_key

    def test_namespace_add(self) -> None:
        self.ns.add('test_key', 1234.5)
        self.assertEqual(self.ns.test_key, 1234.5)
        self.assertEqual(self.ns['test_key'], 1234.5)
        self.assertEqual(self.ts[self.ts.join('ns', 'test_key')], 1234.5)
        with self.assertRaises(KeyError):
            self.ts['test_key']

    def test_delete(self) -> None:
        self.ts.add('test_key', 1234.5)
        self.assertIn('test_key', self.ts)
        self.ts.delete('test_key')
        self.ts.delete('test_key')
        self.assertNotIn('test_key', self.ts)

    def test_namespace_delete(self) -> None:
        self.ts.add('parent_key', 1234.5)
        self.ns.add('child_key', 2345.6)
        self.ns.delete('child_key')
        self.ns.delete('parent_key')
        self.assertEqual(self.ts.keys(), [])

    def test_clear(self) -> None:
        self.ts.add('test_key', 1234.5)
        self.ts.add('test_key_rt', 2345.6)
        self.ts.clear()
        self.assertEqual(self.ts.keys(), [])

    def test_get_default(self) -> None:
        self.assertIsNone(self.ts.get('foo'))
        self.assertEqual(self.ts.get('foo', 'bar'), 'bar')

    def test_get_return_encoded(self) -> None:
        for immutable in [True, False]:
            x = np.array([(1.0, 2), (3.0, 4)], dtype=[('x', float), ('y', int)])
            self.ts.add('test_key_rt', x, immutable=True)
            x_decoded = self.ts.get('test_key_rt')
            self.assertTrue((x_decoded == x).all())
            x_encoded = self.ts.get('test_key_rt', return_encoded=True)
            self.assertEqual(x_encoded, encode_value(x))
            self.ts.delete('test_key_rt')

    def test_get_range_return_encoded(self) -> None:
        test_values = ['Test Value: {}'.format(x) for x in range(5)]
        for i, test_value in enumerate(test_values):
            self.ts.add('test_key', test_value, i)
        stored_values = self.ts.get_range('test_key', st=0)
        self.assertEqual(stored_values[2][0], test_values[2])
        stored_values_pickled = self.ts.get_range('test_key', st=0, return_encoded=True)
        self.assertEqual(stored_values_pickled[2][0], encode_value(test_values[2]))
        # check timestamp
        self.assertEqual(stored_values_pickled[2][1], 2)

    def test_immutable(self) -> None:
        self.ts.add('test_immutable', 1234.5, immutable=True)
        with self.assertRaises(ImmutableKeyError):
            self.ts.add('test_immutable', 1234.5)
        with self.assertRaises(ImmutableKeyError):
            self.ts.add('test_immutable', 5432.1, immutable=True)
        with self.assertRaises(ImmutableKeyError):
            self.ts.set_indexed('test_immutable', 1234.5, 1234.5)

    def test_immutable_same_value(self) -> None:
        self.ts.add('test_immutable', 1234.5, immutable=True)
        self.ts.add('test_mutable', 1234.5)
        with self.assertRaises(ImmutableKeyError):
            self.ts.add('test_mutable', 2345.6, immutable=True)

    @mock.patch('katsdptelstate.encoding._allow_pickle', True)
    @mock.patch('katsdptelstate.encoding._warn_on_pickle', False)
    def test_immutable_same_value_str(self) -> None:
        self.ts.add('test_bytes', b'caf\xc3\xa9', immutable=True)
        self.ts.add('test_bytes', b'caf\xc3\xa9', immutable=True)
        self.ts.add('test_bytes', 'café', immutable=True)
        with self.assertRaises(ImmutableKeyError):
            self.ts.add('test_bytes', b'cafe', immutable=True)
        with self.assertRaises(ImmutableKeyError):
            self.ts.add('test_bytes', 'cafe', immutable=True)
        self.ts.add('test_unicode', 'ümlaut', immutable=True)
        self.ts.add('test_unicode', 'ümlaut', immutable=True)
        self.ts.add('test_unicode', b'\xc3\xbcmlaut', immutable=True)
        with self.assertRaises(ImmutableKeyError):
            self.ts.add('test_unicode', b'umlaut', immutable=True)
        with self.assertRaises(ImmutableKeyError):
            self.ts.add('test_unicode', 'umlaut', immutable=True)
        # Test with a binary string that isn't valid UTF-8
        self.ts.add('test_binary', b'\x00\xff', immutable=True)
        self.ts.add('test_binary', b'\x00\xff', immutable=True)
        # Test Python 2/3 interop by directly injecting the pickled values
        self.ts.backend.set_immutable(b'test_2', b"S'hello'\np1\n.")
        self.ts.backend.set_immutable(b'test_3', b'Vhello\np0\n.')
        self.ts.add('test_2', 'hello', immutable=True)
        self.ts.add('test_3', 'hello', immutable=True)
        # Test handling of the case where the old value cannot be decoded
        # Empty string is never valid encoding
        self.ts.backend.set_immutable(b'test_failed_decode', b'')
        with self.assertRaisesRegex(ImmutableKeyError, 'failed to decode the previous value'):
            self.ts.add('test_failed_decode', '', immutable=True)

    def test_immutable_none(self) -> None:
        self.ts.add('test_none', None, immutable=True)
        self.assertIsNone(self.ts.get('test_none'))
        self.assertIsNone(self.ts.get('test_none', 'not_none'))
        self.assertIsNone(self.ts.test_none)
        self.assertIsNone(self.ts['test_none'])

    def test_immutable_wrong_type(self) -> None:
        self.ts.add('test_mutable', 5)
        self.ts.add('test_indexed', 5, 5)
        with self.assertRaises(ImmutableKeyError):
            self.ts.add('test_mutable', 5, immutable=True)
        with self.assertRaises(ImmutableKeyError):
            self.ts.add('test_indexed', 5, immutable=True)

    def test_namespace_immutable(self) -> None:
        self.ts.add('parent_immutable', 1234.5, immutable=True)
        self.ns.add('child_immutable', 2345.5, immutable=True)
        with self.assertRaises(KeyError):
            self.ts['child_immutable']
        self.assertEqual(self.ns.get('child_immutable'), 2345.5)
        self.assertEqual(self.ns.get('parent_immutable'), 1234.5)

    def test_key_type(self) -> None:
        self.ts.add('parent_immutable', 1, immutable=True)
        self.ns.add('child_immutable', 2, immutable=True)
        self.ts.add('parent', 3)
        self.ns.add('child', 4)
        self.ts.set_indexed('parent_indexed', 'a', 1)
        self.ns.set_indexed('child_indexed', 'b', 2)

        self.assertEqual(self.ts.key_type('parent_immutable'), KeyType.IMMUTABLE)
        self.assertEqual(self.ts.key_type('parent'), KeyType.MUTABLE)
        self.assertEqual(self.ts.key_type('parent_indexed'), KeyType.INDEXED)
        self.assertEqual(self.ts.key_type('child_immutable'), None)
        self.assertEqual(self.ts.key_type('child'), None)
        self.assertEqual(self.ts.key_type('not_a_key'), None)

        self.assertEqual(self.ns.key_type('parent_immutable'), KeyType.IMMUTABLE)
        self.assertEqual(self.ns.key_type('parent'), KeyType.MUTABLE)
        self.assertEqual(self.ns.key_type('parent_indexed'), KeyType.INDEXED)
        self.assertEqual(self.ns.key_type('child_immutable'), KeyType.IMMUTABLE)
        self.assertEqual(self.ns.key_type('child'), KeyType.MUTABLE)
        self.assertEqual(self.ns.key_type('child_indexed'), KeyType.INDEXED)
        self.assertEqual(self.ns.key_type('not_a_key'), None)

    def test_keys(self) -> None:
        self.ts.add('key1', 'a')
        self.ns.add('key2', 'b')
        self.ns.add(b'key2', 'c')
        self.ts.add(b'immutable', 'd', immutable=True)
        self.assertEqual(self.ts.keys(), ['immutable', 'key1', 'ns_key2'])
        self.assertEqual(self.ts.keys('ns_*'), ['ns_key2'])

    def test_complex_store(self) -> None:
        x = np.array([(1.0, 2), (3.0, 4)], dtype=[('x', float), ('y', int)])
        self.ts.add('test_key', x)
        self.assertTrue((self.ts.test_key == x).all())

    def test_contains(self) -> None:
        self.ts.add('test_key', 1234.5)
        self.assertTrue('test_key' in self.ts)
        self.assertFalse('nonexistent_test_key' in self.ts)

    def test_setattr(self) -> None:
        self.ts.test_key = 'foo'
        self.assertEqual(self.ts['test_key'], 'foo')
        self.assertEqual(self.ts.key_type('test_key'), KeyType.IMMUTABLE)
        with self.assertRaises(AttributeError):
            self.ts.root = 'root is a method'    # type: ignore
        self.ts._internal = 'bar'
        self.assertFalse('_internal' in self.ts)

    def test_setitem(self) -> None:
        self.ts['test_key'] = 'foo'
        self.assertEqual(self.ts['test_key'], 'foo')
        self.assertEqual(self.ts.key_type('test_key'), KeyType.IMMUTABLE)

    def test_time_range(self) -> None:
        self.ts.delete('test_key')
        self.ts.add('test_key', 8192, 1)
        self.ts.add('test_key', 16384, 2)
        self.ts.add('test_key', 4096, 3)
        self.ts.add('test_key', 2048, 4)
        self.ts.add('test_immutable', 12345, immutable=True)
        self.assertEqual([(2048, 4)], self.ts.get_range('test_key'))
        self.assertEqual([(16384, 2)], self.ts.get_range('test_key', et=3))
        self.assertEqual([(8192, 1), (16384, 2), (4096, 3)],
                         self.ts.get_range('test_key', st=2, et=4, include_previous=True))
        self.assertEqual([(8192, 1), (16384, 2), (4096, 3), (2048, 4)],
                         self.ts.get_range('test_key', st=0))
        self.assertEqual([(8192, 1), (16384, 2), (4096, 3)],
                         self.ts.get_range('test_key', st=0, et=3.5))
        self.assertEqual([(8192, 1)], self.ts.get_range('test_key', st=-1, et=1.5))
        self.assertEqual([(16384, 2), (4096, 3), (2048, 4)],
                         self.ts.get_range('test_key', st=2))
        self.assertEqual([(8192, 1)], self.ts.get_range('test_key', et=1.5))
        self.assertEqual([], self.ts.get_range('test_key', 3.5, 1.5))
        self.assertEqual([], self.ts.get_range('test_key', et=-0.))
        self.assertEqual([(8192, 1), (16384, 2), (4096, 3), (2048, 4)],
                         self.ts.get_range('test_key', st=1.5, include_previous=True))
        self.assertEqual([(2048, 4)],
                         self.ts.get_range('test_key', st=5, et=6, include_previous=True))
        self.assertEqual([(8192, 1), (16384, 2), (4096, 3)],
                         self.ts.get_range('test_key', st=2, et=4, include_previous=True))
        self.assertRaises(KeyError, self.ts.get_range, 'not_a_key')
        self.assertRaises(ImmutableKeyError, self.ts.get_range, 'test_immutable')

    def test_time_range_include_end(self) -> None:
        self.ts.add('test_key', 1234, 0)
        self.ts.add('test_key', 8192, 1)
        self.ts.add('test_key', 16384, 2)
        self.ts.add('test_key', 4096, 3)
        self.ts.add('test_key', 2048, 4)
        self.assertEqual([], self.ts.get_range('test_key', st=1.5, et=1.5, include_end=True))
        self.assertEqual([(1234, 0)],
                         self.ts.get_range('test_key', st=0.0, et=0.0, include_end=True))
        self.assertEqual([(1234, 0)],
                         self.ts.get_range('test_key', st=0.0, et=-0.0, include_end=True))
        self.assertEqual([(4096, 3), (2048, 4)],
                         self.ts.get_range('test_key', st=3, et=4, include_end=True))
        # include_previous tests
        self.assertEqual([(8192, 1), (16384, 2)],
                         self.ts.get_range('test_key', et=2, include_end=True))
        self.assertEqual([(8192, 1), (16384, 2), (4096, 3)],
                         self.ts.get_range('test_key', st=2, et=3,
                                           include_previous=True, include_end=True))

    def test_add_duplicate(self) -> None:
        self.ts.add('test_key', 'value', 1234.5)
        self.ts.add('test_key', 'value', 1234.5)
        self.assertEqual([('value', 1234.5)], self.ts.get_range('test_key', st=0))

    def test_wait_key_already_done_mutable(self) -> None:
        """Calling wait_key with a condition that is met must return (mutable version)."""
        self.ts.add('test_key', 123)
        value, timestamp = self.ts.get_range('test_key')[0]
        self.ts.wait_key('test_key', lambda v, t: v == value and t == timestamp)

    def test_wait_key_already_done_immutable(self) -> None:
        """Calling wait_key with a condition that is met must return (immutable version)."""
        self.ts.add('test_key', 123, immutable=True)
        self.ts.wait_key('test_key', lambda v, t: v == self.ts['test_key'] and t is None)

    def test_wait_key_already_done_indexed(self) -> None:
        """Calling wait_key with a condition that is met must return (indexed version)."""
        self.ts.set_indexed('test_key', 'idx', 5)
        self.ts.wait_key('test_key', lambda v, t: v == {'idx': 5} and t is None)

    def test_wait_key_timeout(self) -> None:
        """wait_key must time out in the given time if the condition is not met"""
        with self.assertRaises(TimeoutError):
            self.ts.wait_key('test_key', timeout=0.1)
        with self.assertRaises(TimeoutError):
            # Takes a different code path, even though equivalent
            self.ts.wait_key('test_key', lambda value, ts: True, timeout=0.1)

    def _set_key_immutable(self) -> None:
        time.sleep(0.1)
        self.ts['test_key'] = 123

    def _set_key_mutable(self) -> None:
        time.sleep(0.1)
        self.ts.add('test_key', 123, ts=1234567890)

    def _set_key_indexed(self) -> None:
        time.sleep(0.1)
        self.ts.set_indexed('test_key', 'idx', 123)

    def test_wait_key_delayed(self) -> None:
        """wait_key must succeed with a timeout that does not expire before the condition is met."""
        for (set_key, value, timestamp) in [
                (self._set_key_mutable, 123, 1234567890),
                (self._set_key_immutable, 123, None),
                (self._set_key_indexed, {'idx': 123}, None)]:
            thread = threading.Thread(target=set_key)
            thread.start()
            self.ts.wait_key('test_key', lambda v, t: v == value and t == timestamp, timeout=2)
            self.assertEqual(value, self.ts.get('test_key'))
            thread.join()
            self.ts.delete('test_key')

    def test_wait_key_delayed_unconditional(self) -> None:
        """wait_key must succeed when given a timeout that does not expire before key appears."""
        for set_key, value in [
                (self._set_key_mutable, 123),
                (self._set_key_immutable, 123),
                (self._set_key_indexed, {'idx': 123})]:
            thread = threading.Thread(target=set_key)
            thread.start()
            self.ts.wait_key('test_key', timeout=2)
            self.assertEqual(value, self.ts['test_key'])
            thread.join()
            self.ts.delete('test_key')

    def test_wait_key_already_cancelled(self) -> None:
        """wait_key must raise :exc:`CancelledError` if the `cancel_future` is already done."""
        future = mock.MagicMock()
        future.done.return_value = True
        with self.assertRaises(CancelledError):
            self.ts.wait_key('test_key', cancel_future=future)

    def test_wait_key_already_done_and_cancelled(self) -> None:
        """wait_key is successful if both the condition and the cancellation are done on entry."""
        future = mock.MagicMock()
        future.done.return_value = True
        self.ts.add('test_key', 123)
        self.ts.wait_key('test_key', lambda value, ts: value == 123, cancel_future=future)

    def test_wait_key_cancel(self) -> None:
        """wait_key must return when cancelled."""
        def cancel():
            time.sleep(0.1)
            future.done.return_value = True
        future = mock.MagicMock()
        future.done.return_value = False
        thread = threading.Thread(target=cancel)
        thread.start()
        with self.assertRaises(CancelledError):
            self.ts.wait_key('test_key', cancel_future=future)

    def test_wait_key_shadow(self) -> None:
        """updates to a shadowed qualified key must be ignored"""
        def set_key(telstate):
            time.sleep(0.1)
            telstate.add('test_key', True, immutable=True)

        ns2 = self.ns.view('ns2')
        # Put a non-matching key into a mid-level namespace
        self.ns.add('test_key', False)
        # Put a matching key into a shadowed namespace after a delay
        thread = threading.Thread(target=set_key, args=(self.ts,))
        thread.start()
        with self.assertRaises(TimeoutError):
            ns2.wait_key('test_key', lambda value, ts: value is True, timeout=0.5)
        thread.join()

        # Put a matching key into a non-shadowed namespace after a delay
        thread = threading.Thread(target=set_key, args=(ns2,))
        thread.start()
        ns2.wait_key('test_key', lambda value, ts: value is True, timeout=0.5)
        thread.join()

    def test_wait_indexed_already_done(self) -> None:
        self.ts.set_indexed('test_key', 'sub_key', 5)
        self.ts.wait_indexed('test_key', 'sub_key')
        self.ts.wait_indexed('test_key', 'sub_key', lambda v: v == 5)

    def test_wait_indexed_timeout(self) -> None:
        self.ts.set_indexed('test_key', 'sub_key', 5)
        with self.assertRaises(TimeoutError):
            self.ts.wait_indexed('test_key', 'sub_key', lambda v: v == 4, timeout=0.05)
        with self.assertRaises(TimeoutError):
            self.ts.wait_indexed('test_key', 'another_key', timeout=0.05)
        with self.assertRaises(TimeoutError):
            self.ts.wait_indexed('not_present', 'sub_key', timeout=0.05)

    def test_wait_indexed_delayed(self) -> None:
        def set_key():
            self.ts.set_indexed('test_key', 'foo', 1)
            time.sleep(0.05)
            self.ts.set_indexed('test_key', 'bar', 2)
        for condition in [None, lambda value: value == 2]:
            self.ts.delete('test_key')
            thread = threading.Thread(target=set_key)
            thread.start()
            self.ts.wait_indexed('test_key', 'bar', condition, timeout=2)
            self.assertEqual(2, self.ts.get_indexed('test_key', 'bar'))
            thread.join()

    def test_wait_indexed_already_cancelled(self) -> None:
        future = mock.MagicMock()
        future.done.return_value = True
        with self.assertRaises(CancelledError):
            self.ts.wait_indexed('test_key', 'foo', cancel_future=future)

    def test_wait_indexed_already_done_and_cancelled(self) -> None:
        future = mock.MagicMock()
        future.done.return_value = True
        self.ts.set_indexed('test_key', 'sub_key', 1)
        self.ts.wait_indexed('test_key', 'sub_key', lambda value: value == 1, cancel_future=future)

    def test_wait_indexed_cancel(self) -> None:
        def cancel():
            time.sleep(0.1)
            future.done.return_value = True
        future = mock.MagicMock()
        future.done.return_value = False
        thread = threading.Thread(target=cancel)
        thread.start()
        with self.assertRaises(CancelledError):
            self.ts.wait_indexed('test_key', 'sub_key', cancel_future=future)

    def test_wait_indexed_shadow(self) -> None:
        def set_key(telstate):
            time.sleep(0.1)
            telstate.set_indexed('test_key', 'sub_key', 1)

        # Shadow the key that will be set by set_key
        self.ns.set_indexed('test_key', 'blah', 1)
        thread = threading.Thread(target=set_key, args=(self.ts,))
        thread.start()
        with self.assertRaises(TimeoutError):
            self.ns.wait_indexed('test_key', 'sub_key', timeout=0.2)

    def test_wait_indexed_wrong_type(self) -> None:
        def set_key():
            time.sleep(0.1)
            self.ts.add('test_key', 'value')

        thread = threading.Thread(target=set_key)
        thread.start()
        with self.assertRaises(ImmutableKeyError):
            self.ts.wait_indexed('test_key', 'value')
        thread.join()
        # Different code-path when key was already set at the start
        with self.assertRaises(ImmutableKeyError):
            self.ts.wait_indexed('test_key', 'value')

    def _test_mixed_unicode_bytes(self, ns, key) -> None:
        self.ts.clear()
        ns.add(key, 'value', immutable=True)
        self.assertEqual(ns.get(key), 'value')
        self.assertIn(key, ns)
        self.assertEqual(ns.key_type(key), KeyType.IMMUTABLE)
        ns.delete(key)
        ns.add(key, 'value1', ts=1)
        self.assertEqual(
            ns.get_range(key), [('value1', 1.0)])
        ns.wait_key(key)

    def test_mixed_unicode_bytes(self) -> None:
        self._test_mixed_unicode_bytes(self.ts.view(b'ns'), 'test_key')
        self._test_mixed_unicode_bytes(self.ts.view('ns'), b'test_key')

    def test_undecodable_bytes_in_key(self) -> None:
        """Gracefully handle non-UTF-8 bytes in keys."""
        key_b = b'undecodable\xff'
        self.ts.backend.set_immutable(key_b, encode_value('hello'))
        key = [k for k in self.ts.keys() if k.startswith('undecodable')][0]
        self.assertEqual(self.ts.get(key), 'hello')
        self.assertEqual(self.ts.get(key_b), 'hello')
        self.assertEqual(self.ts.get(key_b[1:]), None)

    def test_tab_completion(self) -> None:
        self.ts['z'] = 'value'
        self.ns['b'] = 'value'
        self.ns['a'] = 'value'
        keys = self.ns._ipython_key_completions_()
        self.assertEqual(sorted(keys), ['a', 'b', 'ns_a', 'ns_b', 'z'])
        keys = self.ts._ipython_key_completions_()
        self.assertEqual(sorted(keys), ['ns_a', 'ns_b', 'z'])

    def test_get_indexed(self) -> None:
        self.ts.set_indexed('test_indexed', 'a', 1)
        self.ts.set_indexed('test_indexed', (2, 3j), 2)
        self.assertEqual(self.ts.get_indexed('test_indexed', 'a'), 1)
        self.assertEqual(self.ts.get_indexed('test_indexed', (2, 3j)), 2)
        self.assertIsNone(self.ts.get_indexed('test_indexed', 'missing'))
        self.assertIsNone(self.ts.get_indexed('not_a_key', 'missing'))
        self.assertEqual(self.ts.get_indexed('test_indexed', 'missing', 'default'), 'default')
        self.assertEqual(self.ts.get_indexed('not_a_key', 'missing', 'default'), 'default')
        self.assertEqual(self.ts.get('test_indexed'), {'a': 1, (2, 3j): 2})

    def test_indexed_return_encoded(self) -> None:
        self.ts.set_indexed('test_indexed', 'a', 1)
        values = self.ts.get('test_indexed', return_encoded=True)
        self.assertEqual(values, {
            encode_value('a', encoding=ENCODING_MSGPACK): encode_value(1)
        })
        self.assertEqual(self.ts.get_indexed('test_indexed', 'a', return_encoded=True),
                         encode_value(1))

    def test_set_indexed_immutable(self) -> None:
        self.ts.set_indexed('test_indexed', 'a', 1)
        self.ts.set_indexed('test_indexed', 'a', 1)  # Same value is okay
        with self.assertRaises(ImmutableKeyError):
            self.ts.set_indexed('test_indexed', 'a', 2)
        self.assertEqual(self.ts.get_indexed('test_indexed', 'a'), 1)

    def test_set_indexed_unhashable(self) -> None:
        with self.assertRaises(TypeError):
            self.ts.set_indexed('test_indexed', ['list', 'is', 'not', 'hashable'], 0)

    def test_indexed_wrong_type(self) -> None:
        self.ts['test_immutable'] = 1
        self.ts.add('test_mutable', 2)
        with self.assertRaises(ImmutableKeyError):
            self.ts.set_indexed('test_immutable', 'a', 1)
        with self.assertRaises(ImmutableKeyError):
            self.ts.set_indexed('test_mutable', 'a', 1)
        with self.assertRaises(ImmutableKeyError):
            self.ts.get_indexed('test_immutable', 'a')
        with self.assertRaises(ImmutableKeyError):
            self.ts.get_indexed('test_mutable', 'a')

    def test_namespace_indexed(self) -> None:
        self.ts.set_indexed('test_indexed', 'a', 1)
        self.assertEqual(self.ns.get('test_indexed'), {'a': 1})
        self.assertEqual(self.ns.get_indexed('test_indexed', 'a'), 1)
        self.ns.set_indexed('test_indexed', 'b', 2)
        self.assertEqual(self.ns.get('test_indexed'), {'b': 2})
        self.assertEqual(self.ns.get_indexed('test_indexed', 'b'), 2)
        # Namespace key must completely shadow root
        self.assertIsNone(self.ns.get_indexed('test_indexed', 'a'))


class TestTelescopeStateRedis(TestTelescopeState):
    def make_telescope_state(self) -> TelescopeState:
        def make_fakeredis(**kwargs):
            return fakeredis.FakeRedis()

        with mock.patch('redis.Redis', side_effect=make_fakeredis) as mock_redis:
            ts = TelescopeState('example.com', 1)
            mock_redis.assert_called_with(
                host='example.com',
                db=1, socket_timeout=mock.ANY, health_check_interval=mock.ANY)
        return ts


class TestTelescopeStateRedisUrl(TestTelescopeState):
    def make_telescope_state(self) -> TelescopeState:
        def make_fakeredis(cls, **kwargs):
            return fakeredis.FakeRedis()

        with mock.patch('redis.Redis.from_url', side_effect=make_fakeredis) as mock_redis:
            ts = TelescopeState('redis://example.com', db=1)
            mock_redis.assert_called_with(
                'redis://example.com',
                db=1, socket_timeout=mock.ANY, health_check_interval=mock.ANY)
        return ts



################################################################################
# Copyright (c) 2018-2019, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import logging
import os.path
from typing import Optional, Union, BinaryIO, cast

from rdbtools import RdbParser, RdbCallback

from .errors import RdbParseError
from .utils import _PathType


logger = logging.getLogger(__name__)


class BackendCallback(RdbCallback):
    """A callback adapter that stores keys in backend as RDB file is parsed."""

    def __init__(self) -> None:
        super().__init__(string_escape=None)
        # Counter keeping track of number of keys inserted into backend
        self.n_keys = 0
        # Flag that helps to disambiguate callback errors from parser errors
        self.client_busy = False


def _parse_rdb_file(parser: RdbParser, callback: BackendCallback, fd: BinaryIO,
                    filename: Optional[_PathType] = None) -> None:
    """Apply RDB parser to file descriptor, raising RdbParseError on error."""
    try:
        parser.parse_fd(fd)
    except Exception as exc:
        # Don't remap exception to RdbParseError if it originates from callback
        if callback.client_busy:
            raise
        raise RdbParseError(filename) from exc


def load_from_file(callback: BackendCallback, file: Union[_PathType, BinaryIO]) -> int:
    """Load keys from the specified RDB-compatible dump file into backend.

    Parameters
    ----------
    callback : :class:`katsdptelstate.rdb_reader.BackendCallback`
        Backend-specific callback that stores keys as RDB file is parsed
    file : str or file-like object
        Filename of .rdb file to import, or object representing contents of RDB

    Returns
    -------
    int
        Number of keys loaded into backend

    Raises
    ------
    RdbParseError
        If `file` does not represent a valid RDB file
    """
    size = 'unknown'                      # type: object
    try:
        size = os.path.getsize(cast(_PathType, file))
    except TypeError:
        # One could maybe seek() and tell() on file object but is it worth it?
        pass
    logger.debug("Loading data from RDB dump of %s bytes", size)
    parser = RdbParser(callback)
    try:
        fd = open(cast(_PathType, file), 'rb')
    except TypeError:
        _parse_rdb_file(parser, callback, cast(BinaryIO, file))
    else:
        with fd:
            _parse_rdb_file(parser, callback, fd, cast(_PathType, file))
    return callback.n_keys



################################################################################
# Copyright (c) 2019, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import logging
from typing import List, Tuple, Dict, Union, Optional

from .. import utils
from .backend import Backend
from ..memory import MemoryBackend as SyncMemoryBackend


logger = logging.getLogger(__name__)
_Value = Union[bytes, Dict[bytes, bytes], List[bytes]]


class MemoryBackend(Backend):
    """Telescope state backend that keeps data in memory.

    See :class:`katsdptelstate.memory.MemoryBackend` for details. This class
    is a thin asynchronous wrapper around that version.
    """

    def __init__(self) -> None:
        self._sync = SyncMemoryBackend()

    def to_sync(self) -> SyncMemoryBackend:
        """Get a synchronous backend with the same underlying data."""
        return self._sync

    @staticmethod
    def from_sync(sync: SyncMemoryBackend) -> 'MemoryBackend':
        """Create an asynchronous backend that shares data with a synchronous one."""
        me = MemoryBackend()
        me._sync = sync
        return me

    async def exists(self, key: bytes) -> bool:
        return key in self._sync

    async def keys(self, filter: bytes) -> List[bytes]:
        return self._sync.keys(filter)

    async def delete(self, key: bytes) -> None:
        self._sync.delete(key)

    async def clear(self) -> None:
        self._sync.clear()

    async def key_type(self, key: bytes) -> Optional[utils.KeyType]:
        return self._sync.key_type(key)

    async def set_immutable(self, key: bytes, value: bytes) -> Optional[bytes]:
        return self._sync.set_immutable(key, value)

    async def get(self, key: bytes) -> Union[
            Tuple[None, None],
            Tuple[bytes, None],
            Tuple[bytes, float],
            Tuple[Dict[bytes, bytes], None]]:
        return self._sync.get(key)

    async def add_mutable(self, key: bytes, value: bytes, timestamp: float) -> None:
        self._sync.add_mutable(key, value, timestamp)

    async def set_indexed(self, key: bytes, sub_key: bytes, value: bytes) -> Optional[bytes]:
        return self._sync.set_indexed(key, sub_key, value)

    async def get_indexed(self, key: bytes, sub_key: bytes) -> Optional[bytes]:
        return self._sync.get_indexed(key, sub_key)

    async def get_range(self, key: bytes, start_time: float, end_time: float,
                        include_previous: bool,
                        include_end: bool) -> Optional[List[Tuple[bytes, float]]]:
        return self._sync.get_range(key, start_time, end_time, include_previous, include_end)

    async def dump(self, key: bytes) -> Optional[bytes]:
        return self._sync.dump(key)

    def close(self) -> None:
        self._sync.clear()

    async def wait_closed(self) -> None:
        pass



################################################################################
# Copyright (c) 2015-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Handles details of encoding and decoding strings stored in telescope state."""

import struct
import io
import os
import functools
import threading
import warnings
import pickle
from typing import Optional, Any

import msgpack
import numpy as np

from .errors import EncodeError, DecodeError
from .utils import ensure_binary


PICKLE_PROTOCOL = 2

# Header byte indicating a pickle-encoded value. This is present in pickle
# v2+ (it is the PROTO opcode, which is required to be first). Values below
# 0x80 are assumed to be pickle as well, since v0 uses only ASCII, and v1
# was never used in telstate. Values above 0x80 are reserved for future
# encoding forms.
ENCODING_PICKLE = b'\x80'

#: Header byte indicating a msgpack-encoded value
ENCODING_MSGPACK = b'\xff'

#: Default encoding for :func:`encode_value`
ENCODING_DEFAULT = ENCODING_MSGPACK

#: All encodings that can be used with :func:`encode_value`
ALLOWED_ENCODINGS = frozenset([ENCODING_PICKLE, ENCODING_MSGPACK])

MSGPACK_EXT_TUPLE = 1
MSGPACK_EXT_COMPLEX128 = 2
MSGPACK_EXT_NDARRAY = 3           # .npy format
MSGPACK_EXT_NUMPY_SCALAR = 4      # dtype descriptor then raw value

_allow_pickle = False
_warn_on_pickle = False
_pickle_lock = threading.Lock()       # Protects _allow_pickle, _warn_on_pickle


PICKLE_WARNING = ('The telescope state contains pickled values. This is a security risk, '
                  'but you have enabled it with set_allow_pickle.')
PICKLE_ERROR = ('The telescope state contains pickled values. This is a security risk, '
                'so is disabled by default. If you trust the source of the data, you '
                'can allow the pickles to be loaded by setting '
                'KATSDPTELSTATE_ALLOW_PICKLE=1 in the environment. This is needed for '
                'MeerKAT data up to March 2019.')


# See https://stackoverflow.com/questions/11305790
_pickle_loads = functools.partial(pickle.loads, encoding='latin1')


def set_allow_pickle(allow: bool, warn: bool = False) -> None:
    """Control whether pickles are allowed.

    This overrides the defaults which are determined from the environment.

    Parameters
    ----------
    allow : bool
        If true, allow pickles to be loaded.
    warn : bool
        If true, warn the user the next time a pickle is loaded (after which
        it becomes false). This has no effect if `allow` is false.
    """
    global _allow_pickle, _warn_on_pickle

    with _pickle_lock:
        _allow_pickle = allow
        _warn_on_pickle = warn


def _init_allow_pickle() -> None:
    env = os.environ.get('KATSDPTELSTATE_ALLOW_PICKLE')
    allow = False
    if env == '1':
        allow = True
    elif env == '0':
        allow = False
    elif env is not None:
        warnings.warn('Unknown value {!r} for KATSDPTELSTATE_ALLOW_PICKLE'.format(env))
    set_allow_pickle(allow)


_init_allow_pickle()


def _encode_ndarray(value: np.ndarray) -> bytes:
    fp = io.BytesIO()
    try:
        np.save(fp, value, allow_pickle=False)
    except ValueError as error:
        # Occurs if value has object type
        raise EncodeError(str(error))
    return fp.getvalue()


def _decode_ndarray(data: bytes) -> np.ndarray:
    fp = io.BytesIO(data)
    try:
        return np.load(fp, allow_pickle=False)
    except ValueError as error:
        raise DecodeError(str(error))


def _encode_numpy_scalar(value: np.generic) -> bytes:
    if value.dtype.hasobject:
        raise EncodeError('cannot encode dtype {} as it contains objects'
                          .format(value.dtype))
    descr = np.lib.format.dtype_to_descr(value.dtype)
    return _msgpack_encode(descr) + value.tobytes()


def _decode_numpy_scalar(data: bytes) -> np.generic:
    try:
        descr = _msgpack_decode(data)
        raw = b''
    except msgpack.ExtraData as exc:
        descr = exc.unpacked
        raw = exc.extra
    dtype = np.dtype(descr)
    if dtype.hasobject:
        raise DecodeError('cannot decode dtype {} as it contains objects'
                          .format(dtype))
    value = np.frombuffer(raw, dtype, 1)
    return value[0]


def _msgpack_default(value: Any) -> msgpack.ExtType:
    if isinstance(value, tuple):
        return msgpack.ExtType(MSGPACK_EXT_TUPLE, _msgpack_encode(list(value)))
    elif isinstance(value, np.ndarray):
        return msgpack.ExtType(MSGPACK_EXT_NDARRAY, _encode_ndarray(value))
    elif isinstance(value, np.generic):
        return msgpack.ExtType(MSGPACK_EXT_NUMPY_SCALAR, _encode_numpy_scalar(value))
    elif isinstance(value, complex):
        return msgpack.ExtType(MSGPACK_EXT_COMPLEX128,
                               struct.pack('>dd', value.real, value.imag))
    else:
        raise EncodeError('do not know how to encode type {}'
                          .format(value.__class__.__name__))


def _msgpack_ext_hook(code: int, data: bytes) -> Any:
    if code == MSGPACK_EXT_TUPLE:
        content = _msgpack_decode(data)
        if not isinstance(content, list):
            raise DecodeError('incorrectly encoded tuple')
        return tuple(content)
    elif code == MSGPACK_EXT_COMPLEX128:
        if len(data) != 16:
            raise DecodeError('wrong length for COMPLEX128')
        return complex(*struct.unpack('>dd', data))
    elif code == MSGPACK_EXT_NDARRAY:
        return _decode_ndarray(data)
    elif code == MSGPACK_EXT_NUMPY_SCALAR:
        return _decode_numpy_scalar(data)
    else:
        raise DecodeError('unknown extension type {}'.format(code))


def _msgpack_encode(value: Any) -> bytes:
    return msgpack.packb(value, use_bin_type=True, strict_types=True,
                         default=_msgpack_default)


def _msgpack_decode(value: bytes) -> Any:
    # The max_*_len prevent a corrupted or malicious input from consuming
    # memory significantly in excess of the input size before it determines
    # that there isn't actually enough data to back it.
    max_len = len(value)
    return msgpack.unpackb(value, raw=False, ext_hook=_msgpack_ext_hook,
                           max_str_len=max_len,
                           max_bin_len=max_len,
                           max_array_len=max_len,
                           max_map_len=max_len,
                           max_ext_len=max_len)


def encode_value(value: Any, encoding: bytes = ENCODING_DEFAULT) -> bytes:
    """Encode a value to a byte array for storage in redis.

    Parameters
    ----------
    value
        Value to encode
    encoding
        Encoding method to use, one of the values in :const:`ALLOWED_ENCODINGS`

    Raises
    ------
    ValueError
        If `encoding` is not a recognised encoding
    EncodeError
        EncodeError if the value was not encodable with the chosen encoding.
    """
    if encoding == ENCODING_PICKLE:
        return pickle.dumps(value, protocol=PICKLE_PROTOCOL)
    elif encoding == ENCODING_MSGPACK:
        return ENCODING_MSGPACK + _msgpack_encode(value)
    else:
        raise ValueError('Unknown encoding {:#x}'.format(ord(encoding)))


def decode_value(value: bytes, allow_pickle: Optional[bool] = None) -> Any:
    """Decode a value encoded with :func:`encode_value`.

    The encoded value is self-describing, so it is not necessary to specify
    which encoding was used.

    Parameters
    ----------
    value : bytes
        Encoded value to decode
    allow_pickle : bool, optional
        If false, :const:`ENCODING_PICKLE` is disabled. This is useful for
        security as pickle decoding can execute arbitrary code. If the default
        of ``None`` is used, it is controlled by the
        KATSDPTELSTATE_ALLOW_PICKLE environment variable. If that is not set,
        the default is false. The default may also be overridden with
        :func:`set_allow_pickle`.

    Raises
    ------
    DecodeError
        if `allow_pickle` is false and `value` is a pickle
    DecodeError
        if there was an error in the encoding of `value`
    """
    global _warn_on_pickle

    if not value:
        raise DecodeError('empty value')
    elif value[:1] == ENCODING_MSGPACK:
        try:
            return _msgpack_decode(value[1:])
        except Exception as error:
            raise DecodeError(str(error))
    elif value[:1] <= ENCODING_PICKLE:
        if allow_pickle is None:
            with _pickle_lock:
                allow_pickle = _allow_pickle
                if _warn_on_pickle:
                    warnings.warn(PICKLE_WARNING, FutureWarning)
                    _warn_on_pickle = False
        if allow_pickle:
            try:
                return _pickle_loads(value)
            except Exception as error:
                raise DecodeError(str(error))
        else:
            raise DecodeError(PICKLE_ERROR)
    else:
        raise DecodeError('value starts with unrecognised header byte {!r} '
                          '(katsdptelstate may need to be updated)'.format(value[:1]))


def equal_encoded_values(a: bytes, b: bytes) -> bool:
    """Test whether two encoded values represent the same/equivalent objects.

    This is not a complete implementation. Mostly, it just checks that the
    encoded representation is the same, but to ease the transition to Python 3,
    it will also compare the values if both arguments decode to either
    ``bytes`` or ``str`` (and allows bytes and strings to be equal assuming
    UTF-8 encoding). However, it will not do this recursively in structured
    data.
    """
    if a == b:
        return True
    a = decode_value(a)
    b = decode_value(b)
    try:
        return ensure_binary(a) == ensure_binary(b)
    except TypeError:
        return False



################################################################################
# Copyright (c) 2015-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import unittest
from unittest import mock
from typing import Any

import numpy as np

from ..encoding import encode_value, decode_value, ENCODING_PICKLE, ENCODING_MSGPACK
from ..errors import EncodeError, DecodeError


class _TestEncoding(unittest.TestCase):
    """Test encode_value and decode_value.

    This must be subclassed to specify the encoding type.
    """

    encoding = b''      # Just to keep mypy happy - actual encoding provided by derived classes

    def _test_value(self, value: Any) -> None:
        encoded = encode_value(value, encoding=self.encoding)
        decoded = decode_value(encoded)
        self.assertEqual(type(value), type(decoded))
        if isinstance(value, np.ndarray):
            np.testing.assert_array_equal(value, decoded)
        else:
            self.assertEqual(value, decoded)

    def test_list_tuple(self) -> None:
        self._test_value(('a', 'tuple', ['with', ('embedded', 'list')]))

    def test_float(self) -> None:
        self._test_value(1.0)
        self._test_value(1.23456788901234567890)
        self._test_value(1e300)
        self._test_value(-1e-300)
        self._test_value(float('inf'))
        self._test_value(np.inf)

    def test_simple(self) -> None:
        self._test_value(True)
        self._test_value(False)
        self._test_value(None)

    def test_np_scalar(self) -> None:
        self._test_value(np.float32(5.5))
        self._test_value(np.float64(5.5))
        self._test_value(np.complex128(5.5 + 4.2j))
        self._test_value(np.bool_(True))
        self._test_value(np.bool_(False))
        self._test_value(np.int32(12345678))

    def test_complex(self) -> None:
        self._test_value(1.2 + 3.4j)

    def test_ndarray(self) -> None:
        self._test_value(np.array([1, 2, 3]))

    def test_structured_ndarray(self) -> None:
        dtype = np.dtype([('a', np.int32), ('b', np.int16, (3, 3))])
        self._test_value(np.zeros((2, 3), dtype))

    def test_nan(self) -> None:
        encoded = encode_value(np.nan, encoding=self.encoding)
        decoded = decode_value(encoded)
        self.assertTrue(np.isnan(decoded))

    @mock.patch('katsdptelstate.encoding._allow_pickle', False)
    def test_fuzz(self) -> None:
        if self.encoding == ENCODING_PICKLE:
            raise unittest.SkipTest("Pickles will exhaust memory or crash given a bad pickle")
        # Create an encoded string with a bit of everything
        orig = [('a str', b'bytes'), 3, 4.0, 5 + 6j, np.int32(1),
                True, False, None, np.array([[1, 2, 3]])]
        encoded = encode_value(orig, encoding=self.encoding)
        # Mess with it and check that no exceptions except DecodeError come back
        for i in range(len(encoded)):
            broken = bytearray(encoded)
            for j in range(256):
                broken[i] = j
                try:
                    decode_value(bytes(broken))
                except DecodeError:
                    pass
        # Same thing, but now truncate the message rather than mutating it
        for i in range(len(encoded)):
            try:
                decode_value(encoded[:i])
            except DecodeError:
                pass


@mock.patch('katsdptelstate.encoding._allow_pickle', True)
@mock.patch('katsdptelstate.encoding._warn_on_pickle', False)
class TestEncodingPickle(_TestEncoding):
    encoding = ENCODING_PICKLE


class TestEncodingMsgpack(_TestEncoding):
    encoding = ENCODING_MSGPACK

    def setUp(self) -> None:
        self.object_dtype = np.dtype([('a', np.int32), ('b', np.object_)])
        self.object_array = np.zeros((3,), self.object_dtype)

    def test_ndarray_with_object(self) -> None:
        with self.assertRaises(EncodeError):
            encode_value(self.object_array, encoding=self.encoding)

    def test_numpy_scalar_with_object(self) -> None:
        with self.assertRaises(EncodeError):
            encode_value(self.object_array[0], encoding=self.encoding)

    def test_unhandled_type(self) -> None:
        with self.assertRaises(EncodeError):
            encode_value(self, encoding=self.encoding)



################################################################################
# Copyright (c) 2015-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import logging
from typing import Tuple, Optional, Union, Generic, TypeVar, Any

from .utils import ensure_str, ensure_binary
from .encoding import decode_value, equal_encoded_values
from .errors import ImmutableKeyError, DecodeError


logger = logging.getLogger(__name__)
_B = TypeVar('_B')    # Generic backend
_T = TypeVar('_T', bound='TelescopeStateBase')
_Key = Union[bytes, str]


def check_immutable_change(key: str, old_enc: bytes, new_enc: bytes, new: Any) -> None:
    """Check if an immutable is being changed to the same value.

    If not, raise :exc:`ImmutableKeyError`, otherwise just log a message.
    This is intended to be called by subclasses.

    Parameters
    ----------
    key : str
        Human-readable version of the key
    old_enc : bytes
        Previous value, encoded
    new_enc : bytes
        New value, encoded
    new
        New value, prior to encoding
    """
    try:
        if not equal_encoded_values(new_enc, old_enc):
            raise ImmutableKeyError(
                'Attempt to change value of immutable key {} from '
                '{!r} to {!r}'.format(key, decode_value(old_enc), new))
        else:
            logger.info('Attribute {} updated with the same value'
                        .format(key))
    except DecodeError as error:
        raise ImmutableKeyError(
            'Attempt to set value of immutable key {} to {!r} but '
            'failed to decode the previous value to compare: {}'
            .format(key, new, error)) from error


class TelescopeStateBase(Generic[_B]):
    """Base class for synchronous and asynchronous telescope state classes.

    It contains parts common to :class:`katsdptelstate.TelescopeState` and
    :class:`katsdptelstate.aio.TelescopeState`, mainly relating to setting up
    namespaces.
    """

    SEPARATOR = '_'
    SEPARATOR_BYTES = b'_'

    def __init__(self: _T, backend: Optional[_B] = None,
                 prefixes: Tuple[_Key, ...] = (b'',),
                 base: Optional[_T] = None) -> None:
        if base is not None and backend is not None:
            raise ValueError('cannot specify both base and backend')
        if backend is not None:
            self._backend = backend
        elif base is not None:
            self._backend = base.backend
        else:
            raise ValueError('must specify either base or backend')
        # Ensure all prefixes are bytes internally for consistency
        self._prefixes = tuple(ensure_binary(prefix) for prefix in prefixes)

    @property
    def prefixes(self) -> Tuple[str, ...]:
        """The active key prefixes as a tuple of strings."""
        return tuple(ensure_str(prefix) for prefix in self._prefixes)

    @property
    def backend(self) -> _B:
        return self._backend

    @classmethod
    def join(cls, *names: str) -> str:
        """Join string components of key with supported separator."""
        return cls.SEPARATOR.join(names)

    def view(self: _T, name: _Key, add_separator: bool = True, exclusive: bool = False) -> _T:
        """Create a view with an extra name in the list of namespaces.

        Returns a new view with `name` added as the first prefix, or the
        only prefix if `exclusive` is true. If `name` is non-empty and does not
        end with the separator, it is added (unless `add_separator` is
        false).
        """
        name = ensure_binary(name)
        if name != b'' and name[-1:] != self.SEPARATOR_BYTES and add_separator:
            name += self.SEPARATOR_BYTES
        if exclusive:
            prefixes = (name,)    # type: Tuple[_Key, ...]
        else:
            prefixes = (name,) + self._prefixes
        return self.__class__(prefixes=prefixes, base=self)

    def root(self: _T) -> _T:
        """Create a view containing only the root namespace."""
        return self.__class__(base=self)



################################################################################
# Copyright (c) 2015-2019, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for the Endpoint class."""

from nose.tools import assert_equal, assert_not_equal, assert_raises

from katsdptelstate.endpoint import (
    Endpoint, endpoint_parser, endpoint_list_parser, endpoints_to_str)


class TestEndpoint:
    def test_str(self) -> None:
        assert_equal('test.me:80', str(Endpoint('test.me', 80)))
        assert_equal('[1080::8:800:200C:417A]:12345', str(Endpoint('1080::8:800:200C:417A', 12345)))

    def test_repr(self) -> None:
        assert_equal("Endpoint('test.me', 80)", repr(Endpoint('test.me', 80)))

    def test_parser_default_port(self) -> None:
        parser = endpoint_parser(1234)
        assert_equal(Endpoint('hello', 1234), parser('hello'))
        assert_equal(Endpoint('192.168.0.1', 1234), parser('192.168.0.1'))
        assert_equal(Endpoint('1080::8:800:200C:417A', 1234), parser('[1080::8:800:200C:417A]'))

    def test_parser_port(self) -> None:
        parser = endpoint_parser(1234)
        assert_equal(Endpoint('hello', 80), parser('hello:80'))
        assert_equal(Endpoint('1080::8:800:200C:417A', 80), parser('[1080::8:800:200C:417A]:80'))

    def test_bad_ipv6(self) -> None:
        parser = endpoint_parser(1234)
        assert_raises(ValueError, parser, '[notipv6]:1234')

    def test_iter(self) -> None:
        endpoint = Endpoint('hello', 80)
        assert_equal(('hello', 80), tuple(endpoint))

    def test_eq(self) -> None:
        assert_equal(Endpoint('hello', 80), Endpoint('hello', 80))
        assert_not_equal(Endpoint('hello', 80), Endpoint('hello', 90))
        assert_not_equal(Endpoint('hello', 80), Endpoint('world', 80))
        assert_not_equal(Endpoint('hello', 80), 'not_an_endpoint')

    def test_hash(self) -> None:
        assert_equal(hash(Endpoint('hello', 80)), hash(Endpoint('hello', 80)))
        assert_not_equal(hash(Endpoint('hello', 80)), hash(Endpoint('hello', 90)))


class TestEndpointList:
    def test_parser(self) -> None:
        parser = endpoint_list_parser(1234)
        endpoints = parser(
            'hello:80,world,[1080::8:800:200C:417A],192.168.0.255+4,10.0.255.255+3:60')
        expected = [
            Endpoint('hello', 80),
            Endpoint('world', 1234),
            Endpoint('1080::8:800:200C:417A', 1234),
            Endpoint('192.168.0.255', 1234),
            Endpoint('192.168.1.0', 1234),
            Endpoint('192.168.1.1', 1234),
            Endpoint('192.168.1.2', 1234),
            Endpoint('192.168.1.3', 1234),
            Endpoint('10.0.255.255', 60),
            Endpoint('10.1.0.0', 60),
            Endpoint('10.1.0.1', 60),
            Endpoint('10.1.0.2', 60)
        ]
        assert_equal(expected, endpoints)

    def test_parser_bad_count(self) -> None:
        assert_raises(ValueError, endpoint_list_parser(1234), '192.168.0.1+-4')

    def test_parser_non_integer_count(self) -> None:
        assert_raises(ValueError, endpoint_list_parser(1234), '192.168.0.1+hello')

    def test_parser_count_without_ipv4(self) -> None:
        assert_raises(ValueError, endpoint_list_parser(1234), 'hello.world+4')

    def test_parser_single_port(self) -> None:
        parser = endpoint_list_parser(1234, single_port=True)
        endpoints = parser('hello:1234,world')
        expected = [Endpoint('hello', 1234), Endpoint('world', 1234)]
        assert_equal(expected, endpoints)

    def test_parser_single_port_bad(self) -> None:
        assert_raises(ValueError, endpoint_list_parser(1234, single_port=True), 'x:123,y:456')


def test_endpoints_to_str() -> None:
    endpoints = [
        Endpoint('hostname', 1234),
        Endpoint('1.2.3.4', 7148),
        Endpoint('1.2.3.3', 7148),
        Endpoint('1.2.3.3', None),
        Endpoint('1.2.3.3', 7149),
        Endpoint('1.2.3.5', 7148),
        Endpoint('192.168.1.255', None),
        Endpoint('192.168.2.0', None),
        Endpoint('::10ff', 7148),
        Endpoint('::20ff', 7148),
        Endpoint('::0000:2100', 7148),
        Endpoint('hostname1', None)
    ]
    s = endpoints_to_str(endpoints)
    assert_equal('1.2.3.3,192.168.1.255+1,1.2.3.3+2:7148,1.2.3.3:7149,[::10ff]:7148,'
                 '[::20ff]+1:7148,hostname:1234,hostname1', s)



################################################################################
# Copyright (c) 2019, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import bisect
import re
import logging
from datetime import datetime
from typing import Pattern, List, Tuple, Dict, BinaryIO, Union, Optional

from . import utils
from .utils import _PathType
from .backend import Backend
from .errors import ImmutableKeyError
from .rdb_utility import dump_string, dump_zset, dump_hash
try:
    from . import rdb_reader
    from .rdb_reader import BackendCallback
except ImportError as _rdb_reader_import_error:   # noqa: F841
    rdb_reader = None            # type: ignore
    BackendCallback = object     # type: ignore   # So that MemoryCallback can still be defined


logger = logging.getLogger(__name__)
_Value = Union[bytes, Dict[bytes, bytes], List[bytes]]


def _compile_pattern(pattern: bytes) -> Pattern[bytes]:
    """Compile a glob pattern (e.g. for keys) to a bytes regex.

    fnmatch.fnmatchcase doesn't work for this, because it uses different
    escaping rules to redis, uses ! instead of ^ to negate a character set,
    and handles invalid cases (such as a [ without a ]) differently. This
    implementation was written by studying the redis implementation.
    """
    # This is copy-pasted from fakeredis. It was entirely written by
    # Bruce Merry, so the fakeredis license terms need not apply.

    # It's easier to work with text than bytes, because indexing bytes
    # doesn't behave the same in Python 3. Latin-1 will round-trip safely.
    pattern_str = pattern.decode('latin-1')
    parts = ['^']
    i = 0
    L = len(pattern_str)
    while i < L:
        c = pattern_str[i]
        i += 1
        if c == '?':
            parts.append('.')
        elif c == '*':
            parts.append('.*')
        elif c == '\\':
            if i == L:
                i -= 1
            parts.append(re.escape(pattern_str[i]))
            i += 1
        elif c == '[':
            parts.append('[')
            if i < L and pattern_str[i] == '^':
                i += 1
                parts.append('^')
            parts_len = len(parts)  # To detect if anything was added
            while i < L:
                if pattern_str[i] == '\\' and i + 1 < L:
                    i += 1
                    parts.append(re.escape(pattern_str[i]))
                elif pattern_str[i] == ']':
                    i += 1
                    break
                elif i + 2 < L and pattern_str[i + 1] == '-':
                    start = pattern_str[i]
                    end = pattern_str[i + 2]
                    if start > end:
                        start, end = end, start
                    parts.append(re.escape(start) + '-' + re.escape(end))
                    i += 2
                else:
                    parts.append(re.escape(pattern_str[i]))
                i += 1
            if len(parts) == parts_len:
                if parts[-1] == '[':
                    # Empty group - will never match
                    parts[-1] = '(?:$.)'
                else:
                    # Negated empty group - matches any character
                    assert parts[-1] == '^'
                    parts.pop()
                    parts[-1] = '.'
            else:
                parts.append(']')
        else:
            parts.append(re.escape(c))
    parts.append('\\Z')
    regex = ''.join(parts).encode('latin-1')
    return re.compile(regex, re.S)


class MemoryCallback(BackendCallback):
    """RDB callback that stores keys in :class:`~katsdptelstate.memory.MemoryBackend`."""

    def __init__(self, data: Dict[bytes, _Value]) -> None:
        super().__init__()
        self.data = data

    def set(self, key: bytes, value: bytes, expiry: Optional[datetime], info: dict) -> None:
        self.data[key] = value
        self.n_keys += 1

    def start_sorted_set(self, key: bytes, length: int,
                         expiry: Optional[datetime], info: dict) -> None:
        self.data[key] = []
        self.n_keys += 1

    def zadd(self, key: bytes, score: float, member: bytes) -> None:
        self.data[key].append(member)       # type: ignore

    def end_sorted_set(self, key: bytes) -> None:
        self.data[key].sort()               # type: ignore

    def start_hash(self, key: bytes, length: int, expiry: Optional[datetime], info: dict) -> None:
        self.data[key] = {}
        self.n_keys += 1

    def hset(self, key: bytes, field: bytes, value: bytes) -> None:
        self.data[key][field] = value       # type: ignore


class MemoryBackend(Backend):
    """Telescope state backend that keeps data in memory.

    It is optimised for read-only use, loading data from a .rdb file.
    Write operations are supported only to facilitate testing, but are not
    intended for production use. For that, use a
    :class:`~katsdptelstate.redis.RedisBackend` with an in-memory Redis
    emulation. The :meth:`monitor_keys` method is not implemented.

    Mutable keys are stored as sorted lists, and encode timestamps in-place
    using the same packing as :class:`~katsdptelstate.redis.RedisBackend`.
    """
    def __init__(self) -> None:
        self._data = {}       # type: Dict[bytes, _Value]

    def load_from_file(self, file: Union[_PathType, BinaryIO]) -> int:
        if rdb_reader is None:
            raise _rdb_reader_import_error   # noqa: F821
        return rdb_reader.load_from_file(MemoryCallback(self._data), file)

    def __contains__(self, key: bytes) -> bool:
        return key in self._data

    def keys(self, filter: bytes) -> List[bytes]:
        if filter == b'*':
            return list(self._data.keys())
        else:
            regex = _compile_pattern(filter)
            return [key for key in self._data.keys() if regex.match(key)]

    def delete(self, key: bytes) -> None:
        self._data.pop(key, None)

    def clear(self) -> None:
        self._data.clear()

    def key_type(self, key: bytes) -> Optional[utils.KeyType]:
        value = self._data.get(key)
        if value is None:
            return None
        elif isinstance(value, bytes):
            return utils.KeyType.IMMUTABLE
        elif isinstance(value, dict):
            return utils.KeyType.INDEXED
        elif isinstance(value, list):
            return utils.KeyType.MUTABLE
        else:
            assert False

    def set_immutable(self, key: bytes, value: bytes) -> Optional[bytes]:
        old = self._data.get(key)
        if old is None:
            self._data[key] = value
            return None
        elif isinstance(old, bytes):
            return old
        else:
            raise ImmutableKeyError

    def get(self, key: bytes) -> Union[
            Tuple[None, None],
            Tuple[bytes, None],
            Tuple[bytes, float],
            Tuple[Dict[bytes, bytes], None]]:
        value = self._data.get(key)
        if isinstance(value, list):
            return utils.split_timestamp(value[-1])
        else:
            # mypy is not smart enough to figure out that this is compatible
            return value, None     # type: ignore

    def add_mutable(self, key: bytes, value: bytes, timestamp: float) -> None:
        str_val = utils.pack_timestamp(timestamp) + value
        items = self._data.get(key)
        if items is None:
            self._data[key] = [str_val]
        elif isinstance(items, list):
            # To match redis behaviour, we need to avoid inserting the item
            # if it already exists.
            pos = bisect.bisect_left(items, str_val)
            if pos == len(items) or items[pos] != str_val:
                items.insert(pos, str_val)
        else:
            raise ImmutableKeyError

    def set_indexed(self, key: bytes, sub_key: bytes, value: bytes) -> Optional[bytes]:
        item = self._data.setdefault(key, {})
        if not isinstance(item, dict):
            raise ImmutableKeyError
        if sub_key in item:
            return item[sub_key]
        else:
            item[sub_key] = value
            return None

    def get_indexed(self, key: bytes, sub_key: bytes) -> Optional[bytes]:
        item = self._data[key]
        if not isinstance(item, dict):
            raise ImmutableKeyError
        return item.get(sub_key)

    @classmethod
    def _bisect(cls, items: List[bytes], timestamp: float,
                is_end: bool, include_end: bool = False) -> int:
        packed = utils.pack_query_timestamp(timestamp, is_end, include_end)
        if packed == b'-':
            return 0
        elif packed == b'+':
            return len(items)
        else:
            # pack_query_timestamp adds a prefix of [ or (, which we don't need
            return bisect.bisect_left(items, packed[1:])

    def get_range(self, key: bytes, start_time: float, end_time: float,
                  include_previous: bool, include_end: bool) -> Optional[List[Tuple[bytes, float]]]:
        items = self._data.get(key)
        if items is None:
            return None
        elif not isinstance(items, list):
            raise ImmutableKeyError

        start_pos = self._bisect(items, start_time, False)
        if include_previous and start_pos > 0:
            start_pos -= 1
        end_pos = self._bisect(items, end_time, True, include_end)
        return [utils.split_timestamp(value) for value in items[start_pos:end_pos]]

    def dump(self, key: bytes) -> Optional[bytes]:
        value = self._data.get(key)
        if value is None:
            return None
        elif isinstance(value, bytes):
            return dump_string(value)
        elif isinstance(value, list):
            return dump_zset(value)
        elif isinstance(value, dict):
            return dump_hash(value)
        else:
            assert False



################################################################################
# Copyright (c) 2015-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

from abc import ABC, abstractmethod
import asyncio
import time
from typing import List, Tuple, Dict, Iterable, AsyncGenerator, Optional, Union

from ..utils import KeyType
from ..backend import KeyUpdateBase


class Backend(ABC):
    """Low-level interface for telescope state backends.

    The backend interface does not deal with namespaces or encodings, which are
    handled by the frontend :class:`TelescopeState` class. A backend must be
    able to store the same types as :class:`~katsdptelstate.TelescopeState`,
    but keys and values will be :class:`bytes` rather than arbitrary Python
    objects.
    """

    @abstractmethod
    async def exists(self, key: bytes) -> bool:
        """Return if `key` is in the backend."""

    @abstractmethod
    async def keys(self, filter: bytes) -> List[bytes]:
        """Return all keys matching `filter`.

        The filter is a redis pattern. Backends might only support ``b'*'`` as
        a filter.
        """

    @abstractmethod
    async def delete(self, key: bytes) -> None:
        """Delete a key (no-op if it does not exist)"""

    @abstractmethod
    async def clear(self) -> None:
        """Remove all keys"""

    @abstractmethod
    async def key_type(self, key: bytes) -> Optional[KeyType]:
        """Get type of `key`, or ``None`` if it does not exist."""

    @abstractmethod
    async def set_immutable(self, key: bytes, value: bytes) -> Optional[bytes]:
        """Set the value of an immutable key.

        If the key already exists (and is immutable), returns the existing
        value and does not update it. Otherwise, returns ``None``.

        Raises
        ------
        ImmutableKeyError
            If the key exists and is not immutable.
        """

    @abstractmethod
    async def get(self, key: bytes) -> Union[
            Tuple[None, None],
            Tuple[bytes, None],
            Tuple[bytes, float],
            Tuple[Dict[bytes, bytes], None]]:
        """Get the value and timestamp of a key.

        The return value depends on the key type:

        immutable
          The value.
        mutable
          The most recent value.
        indexed
          A dictionary of all values (with undefined iteration order).
        absent
          None

        The timestamp will be ``None`` for types other than mutable.
        """

    @abstractmethod
    async def add_mutable(self, key: bytes, value: bytes, timestamp: float) -> None:
        """Set a (value, timestamp) pair in a mutable key.

        The `timestamp` will be a non-negative float value.

        Raises
        ------
        ImmutableKeyError
            If the key exists and is not mutable
        """

    @abstractmethod
    async def set_indexed(self, key: bytes, sub_key: bytes, value: bytes) -> Optional[bytes]:
        """Add value in an indexed immutable key.

        If the sub-key already exists, returns the existing value and does not
        update it. Otherwise, returns ``None``.

        Raises
        ------
        ImmutableKeyError
            If the key exists and is not indexed.
        """

    @abstractmethod
    async def get_indexed(self, key: bytes, sub_key: bytes) -> Optional[bytes]:
        """Get the value of an indexed immutable key.

        Returns ``None`` if the key exists but the sub-key does not exist.

        Raises
        ------
        KeyError
            If the key does not exist.
        ImmutableKeyError
            If the key exists and is not indexed.
        """

    @abstractmethod
    async def get_range(self, key: bytes, start_time: float, end_time: float,
                        include_previous: bool,
                        include_end: bool) -> Optional[List[Tuple[bytes, float]]]:
        """Obtain a range of values from a mutable key.

        If the key does not exist, returns None.

        Parameters
        ----------
        key : bytes
            Key to search
        start_time : float
            Start of the range (inclusive).
        end_time : float
            End of the range. It is guaranteed to be non-negative.
        include_previous : bool
            If true, also return the last entry prior to `start_time`.
        include_end : bool
            If true, treat `end_time` as inclusive, otherwise exclusive.

        Raises
        ------
        ImmutableKeyError
            If the key exists and is not mutable
        """

    @abstractmethod
    async def dump(self, key: bytes) -> Optional[bytes]:
        """Return a key in the same format as the Redis DUMP command, or None if not present."""

    async def monitor_keys(self, keys: Iterable[bytes]) -> AsyncGenerator[KeyUpdateBase, None]:
        """Report changes to keys in `keys`.

        Returns an asynchronous iterator that yields an infinite stream of
        update notifications. When no longer needed it should be closed.
        """
        # This is a valid but usually suboptimal implementation
        while True:
            await asyncio.sleep(1)
            yield KeyUpdateBase()
        timeout = yield None
        while True:
            assert timeout is not None
            time.sleep(timeout)
            yield KeyUpdateBase()

    @abstractmethod
    def close(self) -> None:
        """Start shutting down the connection to the backing storage."""

    @abstractmethod
    async def wait_closed(self) -> None:
        """Wait until the shutdown started by :meth:`close` has completed."""



################################################################################
# Copyright (c) 2019-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import asyncio
import contextlib
import hashlib
import logging
from typing import (List, Tuple, Dict, Generator, AsyncGenerator, Sequence,
                    Iterable, Callable, Awaitable, Optional, Union, Any)

import pkg_resources
import aioredis

from .. import utils
from .backend import Backend
from ..backend import KeyUpdateBase, MutableKeyUpdate, ImmutableKeyUpdate, IndexedKeyUpdate
from ..errors import ImmutableKeyError
from ..utils import KeyType, ensure_str, display_str


logger = logging.getLogger(__name__)


@contextlib.contextmanager
def _handle_wrongtype() -> Generator[None, None, None]:
    """Map redis WRONGTYPE error to ImmutableKeyError.

    It also handles WRONGTYPE errors from inside scripts.
    """
    try:
        yield
    except aioredis.ReplyError as error:
        if (error.args[0].startswith('WRONGTYPE ')
                or (error.args[0].startswith('ERR Error running script')
                    and 'WRONGTYPE ' in error.args[0])):
            raise ImmutableKeyError
        raise


class _Script:
    """Handle caching of scripts on the redis server."""

    def __init__(self, script: bytes) -> None:
        self.script = script
        self.digest = hashlib.sha1(self.script).hexdigest()

    async def __call__(self, client: aioredis.Redis,
                       keys: Sequence = [], args: Sequence = []) -> Any:
        try:
            return await client.evalsha(self.digest, keys, args)
        except aioredis.ReplyError as exc:
            if not exc.args[0].startswith('NOSCRIPT '):
                raise
        # The script wasn't cached, so try again with original source
        return await client.eval(self.script, keys, args)


class RedisBackend(Backend):
    """Backend for :class:`TelescopeState` using redis for storage."""

    _KEY_TYPES = {
        b'none': None,
        b'string': KeyType.IMMUTABLE,
        b'hash': KeyType.INDEXED,
        b'zset': KeyType.MUTABLE
    }

    def __init__(self, client: aioredis.Redis) -> None:
        self.client = client
        self._scripts = {}     # type: Dict[str, _Script]
        for script_name in ['get', 'set_immutable', 'get_indexed', 'set_indexed',
                            'add_mutable', 'get_range']:
            script = pkg_resources.resource_string(
                'katsdptelstate', 'lua_scripts/{}.lua'.format(script_name))
            self._scripts[script_name] = _Script(script)

    async def _execute(self, call: Callable[..., Awaitable], *args, **kwargs) -> Any:
        """Execute a redis command, retrying if the connection died.

        This handles cases like an idle connection being silently closed,
        the server restarting etc. If there is a connection error, a fresh
        connection is made and one more attempt is made.

        Since there is no way to tell if the original attempt actually
        succeeded, the command must be idempotent.
        """
        try:
            return await call(*args, **kwargs)
        except (ConnectionError, aioredis.ConnectionClosedError) as exc:
            # Closes all connections in the pool to ensure that we get a
            # fresh connection.
            logger.warning('redis connection error (%s), trying again', exc)
            await self.client.connection.clear()
            return await call(*args, **kwargs)

    async def _call(self, script_name: str, *args, **kwargs) -> Any:
        """Call a Lua script by name.

        This uses _execute, so the script must be idempotent.
        """
        return await self._execute(self._scripts[script_name], self.client, *args, **kwargs)

    async def exists(self, key: bytes) -> bool:
        return bool(await self._execute(self.client.exists, key))

    async def keys(self, filter: bytes) -> List[bytes]:
        return await self._execute(self.client.keys, filter)

    async def delete(self, key: bytes) -> None:
        await self._execute(self.client.delete, key)

    async def clear(self) -> None:
        await self._execute(self.client.flushdb)

    async def key_type(self, key: bytes) -> Optional[KeyType]:
        type_ = await self._execute(self.client.type, key)
        return self._KEY_TYPES[type_]

    async def set_immutable(self, key: bytes, value: bytes) -> Optional[bytes]:
        with _handle_wrongtype():
            return await self._call('set_immutable', [key], [value])

    async def get(self, key: bytes) -> Union[
            Tuple[None, None],
            Tuple[bytes, None],
            Tuple[bytes, float],
            Tuple[Dict[bytes, bytes], None]]:
        result = await self._call('get', [key])
        if result[1] == b'none':
            return None, None
        elif result[1] == b'string':
            return result[0], None
        elif result[1] == b'zset':
            return utils.split_timestamp(result[0])
        elif result[1] == b'hash':
            it = iter(result[0])
            d = dict(zip(it, it))
            return d, None
        else:
            raise RuntimeError('Unknown key type {} for key {}'
                               .format(ensure_str(result[1]), display_str(key)))

    async def add_mutable(self, key: bytes, value: bytes, timestamp: float) -> None:
        str_val = utils.pack_timestamp(timestamp) + value
        with _handle_wrongtype():
            await self._call('add_mutable', [key], [str_val])

    async def set_indexed(self, key: bytes, sub_key: bytes, value: bytes) -> Optional[bytes]:
        with _handle_wrongtype():
            return await self._call('set_indexed', [key], [sub_key, value])

    async def get_indexed(self, key: bytes, sub_key: bytes) -> Optional[bytes]:
        with _handle_wrongtype():
            result = await self._call('get_indexed', [key], [sub_key])
            if result == 1:
                # Key does not exist
                raise KeyError
            else:
                return result

    async def get_range(self, key: bytes, start_time: float, end_time: float,
                        include_previous: bool,
                        include_end: bool) -> Optional[List[Tuple[bytes, float]]]:
        packed_st = utils.pack_query_timestamp(start_time, False)
        packed_et = utils.pack_query_timestamp(end_time, True, include_end)
        with _handle_wrongtype():
            ret_vals = await self._call('get_range', [key],
                                        [packed_st, packed_et, int(include_previous)])
        if ret_vals is None:
            return None     # Key does not exist
        return [utils.split_timestamp(val) for val in ret_vals]

    async def monitor_keys(self, keys: Iterable[bytes]) -> AsyncGenerator[KeyUpdateBase, None]:
        # Refer to katsdptelstate.redis.RedisBackend for details of the protocol.
        while True:
            receiver = aioredis.pubsub.Receiver()
            # We use a dedicated connection for the lifetime of this generator.
            # While the connection pool also has a member to hold a pubsub
            # connection, it's more robust to just close a connection than to
            # try to manage unsubscriptions.
            client: Optional[aioredis.Redis] = None
            try:
                conn = await self.client.connection.acquire()
                client = aioredis.Redis(conn)
                await client.subscribe(*(receiver.channel(b'update/' + key) for key in keys))
                # Condition may have been satisfied while we were busy subscribing
                logger.debug('Subscribed to channels, notifying caller to check')
                yield KeyUpdateBase()
                async for channel, data in receiver.iter():
                    assert channel.name.startswith(b'update/')
                    key = channel.name[7:]
                    logger.debug('Received update on channel %r', key)
                    # First check if there is a key type byte in the message.
                    try:
                        key_type = KeyType(data[0])
                    except (ValueError, IndexError):
                        # It's the older format lacking a key type byte
                        key_type2 = await self.key_type(key)
                        if key_type2 is None:
                            continue     # Key was deleted before we could retrieve the type
                        key_type = key_type2
                    else:
                        data = data[1:]  # Strip off the key type
                    if key_type == KeyType.IMMUTABLE:
                        yield ImmutableKeyUpdate(key, data)
                    elif key_type == KeyType.INDEXED:
                        # Encoding is <sub-key-length>\n<sub-key><value>
                        newline = data.find(b'\n')
                        if newline == -1:
                            raise RuntimeError('Pubsub message missing newline')
                        sub_key_len = int(data[:newline])
                        sub_key = data[newline + 1 : newline + 1 + sub_key_len]
                        value = data[newline + 1 + sub_key_len :]
                        yield IndexedKeyUpdate(key, sub_key, value)
                    elif key_type == KeyType.MUTABLE:
                        value, timestamp = utils.split_timestamp(data)
                        yield MutableKeyUpdate(key, value, timestamp)
                    else:
                        raise RuntimeError('Unhandled key type {}'.format(key_type))
            except (ConnectionError, aioredis.ConnectionClosedError) as exc:
                logger.warning('pubsub connection error (%s), retrying in 1s', exc)
            finally:
                if client is not None:
                    client.close()
                    await client.wait_closed()
                receiver.stop()
            await asyncio.sleep(1)

    async def dump(self, key: bytes) -> Optional[bytes]:
        return await self._execute(self.client.dump, key)

    def close(self) -> None:
        self.client.close()

    async def wait_closed(self) -> None:
        await self.client.wait_closed()



################################################################################
# Copyright (c) 2018-2019, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Implements encoding to the RDB file format.

Some documentation is included here, but for a more complete reference, see
https://github.com/sripathikrishnan/redis-rdb-tools/wiki/Redis-RDB-Dump-File-Format
"""

import struct
import itertools
from typing import Mapping, Sequence, Iterable, Union


# Version 6 (first two bytes), and a zero checksum
DUMP_POSTFIX = b"\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00"


def encode_len(length: int) -> bytes:
    """Encodes the specified length as 1,2 or 5 bytes of
    RDB specific length encoded byte.

    - For values less than 64 (i.e two MSBs zero - encode directly in the byte)

    - For values less than 16384 use two bytes, leading MSBs are 01 followed by
      14 bits encoding the value

    - For values less than (2^32 - 1) use 5 bytes, leading MSBs are 10. Length
      encoded only in the lowest 32 bits.
    """
    if length > (2**32 - 1):
        raise ValueError("Cannot encode item of length {} as it is greater than 2^32 -1"
                         .format(length))
    if length < 64:
        return struct.pack('B', length)
    if length < 16384:
        return struct.pack(">H", 0x4000 + length)
    return struct.pack('>Q', 0x8000000000 + length)[3:]


def encode_prev_length(length: int) -> bytes:
    """Special helper for zset previous entry lengths.

    If length < 253 then use 1 byte directly, otherwise
    set first byte to 254 and add 4 trailing bytes as an
    unsigned integer.
    """
    if length < 254:
        return struct.pack('B', length)
    return b'\xfe' + struct.pack("<I", length)


def dump_string(data: bytes) -> bytes:
    """Encode a binary string as per redis DUMP command"""
    type_specifier = b'\x00'
    encoded_length = encode_len(len(data))
    return type_specifier + encoded_length + data + DUMP_POSTFIX


def encode_ziplist(entries: Iterable[Union[bytes, int]]) -> bytes:
    """Create an RDB ziplist, including the envelope.

    The entries can either be byte strings or integers, and any iterable can
    be used.

    The ziplist string envelope itself is an RDB-encoded string with the
    following form:
    (format descriptions are provided in the description for each component)

        (bytes in list '<i')(offset to tail '<i')
        (number of entries '<h')(entry 1)(entry 2)(entry N)(terminator 0xFF)

    The entries themselves are encoded as follows:

        (previous entry length 1byte or 5bytes)
        (entry length - up to 5 bytes as per length encoding)(value)
    """
    def append(raw):
        nonlocal zl_len
        raw_entries.append(raw)
        zl_len += len(raw)

    raw_entries = [b'']   # Later replaced by a header
    zl_len = 10           # Header is 10 bytes
    zl_entries = 0
    zl_last = zl_len      # Offset to previous entry
    for entry in entries:
        previous_length = zl_len - zl_last
        zl_last = zl_len
        zl_entries += 1
        append(encode_prev_length(previous_length))
        if isinstance(entry, int):
            # Special encoding for integers. At present only the value 0 is
            # used, so we only implement the 4-bit encoding, and otherwise
            # fall back to string form.
            if 0 <= entry <= 12:
                append(struct.pack('<B', 0xf1 + entry))
                continue
            else:
                entry = b'%d' % entry
        assert isinstance(entry, bytes)
        append(encode_len(len(entry)))
        append(entry)
    append(b'\xff')
    # Fill in the header
    raw_entries[0] = (struct.pack('<I', zl_len) + struct.pack('<I', zl_last)
                      + struct.pack('<H', zl_entries))
    return b''.join(raw_entries)


def dump_iterable(prefix: bytes, suffix: bytes, entries: Iterable[bytes]) -> bytes:
    """Output a sequence of strings, bracketed by a prefix and suffix.

    This is an internal function used to implement :func:`dump_zset` and
    :func:`dump_hash`.
    """
    enc = [prefix]
    for entry in entries:
        enc.extend([encode_len(len(entry)), entry])
    enc.append(suffix)
    return b''.join(enc)


def dump_zset(data: Sequence[bytes]) -> bytes:
    """Encode a set of values as a redis zset, as per redis DUMP command.

    All scores are assumed to be zero and encoded in the most efficient way
    possible.

    Redis uses both LZF and Ziplist format depending on various arcane heuristics.
    For maximum compatibility we use Ziplist for small lists.

    There are two possible ways to encode the zset:

      1. Using a ziplist: (0c)(ziplist) where (ziplist) is string-encoded.
      2. Using a simple list (03)(num entries)(entry)(entry)...
    """
    entry_count = 2 * len(data)

    # The entries counter for ziplists is encoded as 2 bytes, if we exceed this limit
    # we fall back to making a simple set. Redis of course makes the decision point using
    # only 7 out of the 16 bits available and switches at 127 entries...
    # Additionally, ziplists can't be used if the total size is too large to encode. For
    # safety, we switch at 2GB rather than risking off-by-one errors.
    if entry_count > 127 or sum(len(entry) for entry in data) >= 2**31:
        # for inscrutable reasons the length here is half the number of actual
        # entries (i.e. scores are ignored)
        return dump_iterable(
            b'\x03' + encode_len(len(data)),
            DUMP_POSTFIX,
            # interleave entries and scores directly
            # for now, scores are kept fixed at integer zero since float
            # encoding is breaking for some reason and is not needed for telstate
            itertools.chain.from_iterable((entry, b'0') for entry in data)
        )
    else:
        zl_envelope = encode_ziplist(itertools.chain.from_iterable((x, 0) for x in data))
        return b'\x0c' + encode_len(len(zl_envelope)) + zl_envelope + DUMP_POSTFIX


def dump_hash(data: Mapping[bytes, bytes]) -> bytes:
    """Encode a dictionary as a redis hash, as per redis DUMP command.

    The encoding is similar to zsets, but using the value instead of the score.
    """
    entry_count = 2 * len(data)

    if entry_count > 512 or sum(len(entry[0]) + len(entry[1]) for entry in data.items()) >= 2**31:
        return dump_iterable(
            b'\x04' + encode_len(len(data)),
            DUMP_POSTFIX,
            itertools.chain.from_iterable(data.items())
        )
    else:
        zl_envelope = encode_ziplist(itertools.chain.from_iterable(data.items()))
        return b'\x0d' + encode_len(len(zl_envelope)) + zl_envelope + DUMP_POSTFIX



################################################################################
# Copyright (c) 2018-2019, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

"""Tests for the RDB handling (reading and writing) functionality."""

import unittest
import shutil
import os
import tempfile
from typing import Mapping, Iterable, BinaryIO, Optional, Union

import redis
import fakeredis

from katsdptelstate.rdb_writer import RDBWriter
from katsdptelstate.rdb_reader import load_from_file
from katsdptelstate.rdb_utility import dump_string, dump_zset, dump_hash
from katsdptelstate.redis import RedisBackend, RedisCallback
from katsdptelstate import TelescopeState, RdbParseError, KeyType


class TabloidRedis(fakeredis.FakeRedis):
    """A Redis-like class that is a very superficial simulacrum of a real server.

    Designed specifically to support the read cases in use by katsdptelstate.
    The Redis-like functionality is almost entirely derived from FakeRedis,
    we only add a dump function.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def dump(self, key: Union[bytes, str]) -> Optional[bytes]:
        """Encode Redis key value in an RDB compatible format.

        Note: This follows the DUMP command in Redis itself which produces
        output that is similarly encoded to an RDB, but not exactly the same.

        ZSet scores are ignored and encoded as zero.

        Returns None if `key` not found.
        """
        key_type = self.type(key)
        if key_type == b'none':
            return None
        elif key_type == b'zset':
            data = self.zrange(key, 0, -1)
            return dump_zset(data)
        elif key_type == b'string':
            return dump_string(self.get(key))
        elif key_type == b'hash':
            return dump_hash(self.hgetall(key))
        raise NotImplementedError("Unsupported key type {}. Must be either "
                                  "string or zset".format(key_type))


class TestRDBHandling(unittest.TestCase):
    """Test that data can be written by :class:`~.TabloidRedis` then read back"""
    def setUp(self) -> None:
        # an empty tabloid redis instance
        self.tr = TabloidRedis()
        self.base_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.base_dir)

    def base(self, filename: str) -> str:
        return os.path.join(self.base_dir, filename)

    def _add_test_vec(self, key: str) -> None:
        # TabloidRedis does not properly support non-zero scores
        self.tr.zadd(key, {b'first': 0.0, b'second': 0.0, b'third\n\0': 0.0})

    def test_writer_reader(self) -> None:
        self._add_test_vec('writezl')
        test_str = b"some string\x00\xa3\x17\x43and now valid\xff"
        self.tr.set('write', test_str)
        local_tr = TabloidRedis()
        local_tr.set('extra', test_str)

        with RDBWriter(self.base('all.rdb')) as rdbw:
            rdbw.save(self.tr)
            rdbw.save(local_tr)
        self.assertEqual(rdbw.keys_written, 3)
        self.assertEqual(rdbw.keys_failed, 0)
        # Also test that RDBWriter can work without a with-statement
        rdbw = RDBWriter(self.base('one.rdb'))
        rdbw.save(self.tr, keys=['writezl'])
        rdbw.close()
        self.assertEqual(rdbw.keys_written, 1)
        self.assertEqual(rdbw.keys_failed, 0)
        with RDBWriter(self.base('broken.rdb')) as rdbw:
            rdbw.save(self.tr, keys=['does_not_exist'])
        self.assertEqual(rdbw.keys_written, 0)
        self.assertEqual(rdbw.keys_failed, 1)

        local_tr = TabloidRedis()
        self.assertEqual(load_from_file(RedisCallback(local_tr), self.base('all.rdb')), 3)
        self.assertEqual(set(local_tr.keys()), {b'write', b'writezl', b'extra'})
        self.assertEqual(local_tr.get('write'), test_str)
        self.assertEqual(local_tr.get('extra'), test_str)
        vec = local_tr.zrange('writezl', 0, -1, withscores=True)
        self.assertEqual(vec, [(b'first', 0.0), (b'second', 0.0), (b'third\n\0', 0.0)])

        local_tr = TabloidRedis()
        load_from_file(RedisCallback(local_tr), self.base('one.rdb'))
        self.assertEqual(local_tr.keys(), [b'writezl'])
        vec = local_tr.zrange('writezl', 0, -1, withscores=True)
        self.assertEqual(vec, [(b'first', 0.0), (b'second', 0.0), (b'third\n\0', 0.0)])

    def _test_zset(self, items: Iterable[bytes]) -> None:
        self.tr.zadd('my_zset', {x: 0.0 for x in items})
        with RDBWriter(self.base('zset.rdb')) as rdbw:
            rdbw.save(self.tr)

        local_tr = TabloidRedis()
        load_from_file(RedisCallback(local_tr), self.base('zset.rdb'))
        self.assertEqual(local_tr.keys(), [b'my_zset'])
        vec = local_tr.zrange('my_zset', 0, -1, withscores=True)
        self.assertEqual(vec, [(item, 0.0) for item in items])

    def test_zset_many_entries(self) -> None:
        """Zset with more than 127 entries.

        This uses the more general encoding, rather than ziplist.
        """
        self._test_zset([b'item%03d' % i for i in range(200)])

    def test_zset_with_big_entry(self) -> None:
        """Ziplist with large entry (has different encoding)"""
        self._test_zset([b'?' * 100000])

    # Disabled because it uses too much memory in Jenkins
    # def test_zset_4gb(self) -> None:
    #     """Ziplist with >4GB of data (can't be encoded as ziplist)"""
    #     self._test_zset([(b'%03d' % i) + b'?' * 500000000 for i in range(10)])

    def _test_hash(self, items: Mapping[bytes, bytes]) -> None:
        self.tr.hmset('my_hash', items)
        with RDBWriter(self.base('hash.rdb')) as rdbw:
            rdbw.save(self.tr)

        local_tr = TabloidRedis()
        load_from_file(RedisCallback(local_tr), self.base('hash.rdb'))
        self.assertEqual(local_tr.keys(), [b'my_hash'])
        read = local_tr.hgetall('my_hash')
        self.assertEqual(read, items)

    def test_hash_many_entries(self) -> None:
        """Hash with large number of entries.

        This uses the more general encoding, rather than a ziplist.
        """
        self._test_hash({b'key%03d' % i: b'value%03d' % i for i in range(1000)})

    def test_hash_big_entry(self) -> None:
        """Ziplist with a large entry (has different encoding)"""
        self._test_hash({b'x' * 1000000: b'y' * 100000})


class TestLoadFromFile(unittest.TestCase):
    """Test :meth:`TelescopeState.load_from_file`."""

    def setUp(self) -> None:
        # an empty tabloid redis instance
        self.base_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.base_dir)
        self.filename = os.path.join(self.base_dir, 'dump.rdb')

    def make_telescope_state(self) -> TelescopeState:
        return TelescopeState()

    def save_to_file(self, file: str) -> None:
        write_ts = self.make_telescope_state()
        write_ts['immutable'] = ['some value']
        write_ts.add('mutable', 'first', 12.0)
        write_ts.add('mutable', 'second', 15.5)
        write_ts.set_indexed('indexed', 'a', 1j)
        write_ts.set_indexed('indexed', 'b', 'z')
        # Write data to file
        with RDBWriter(file) as rdbw:
            rdbw.save(write_ts)

    def load_from_file_and_check(self, file: Union[str, BinaryIO]) -> None:
        # Load RDB file back into some backend
        read_ts = self.make_telescope_state()
        read_ts.load_from_file(file)
        self.assertEqual(read_ts.keys(), ['immutable', 'indexed', 'mutable'])
        self.assertEqual(read_ts.key_type('immutable'), KeyType.IMMUTABLE)
        self.assertEqual(read_ts['immutable'], ['some value'])
        self.assertEqual(read_ts.get_range('mutable', st=0),
                         [('first', 12.0), ('second', 15.5)])
        self.assertEqual(read_ts.get('indexed'), {'a': 1j, 'b': 'z'})

    def test_load_from_file(self) -> None:
        self.save_to_file(self.filename)
        # Check loading from filenames and file-like objects
        self.load_from_file_and_check(self.filename)
        self.load_from_file_and_check(open(self.filename, 'rb'))
        # Check that malformed RDB file raises the appropriate exception
        file = open(self.filename, 'rb')
        file.read(1)
        with self.assertRaises(RdbParseError):
            self.load_from_file_and_check(file)
        # Check what happens if file does not exist
        with self.assertRaises(OSError):
            self.load_from_file_and_check(self.filename + '.nonexistent')


class TestLoadFromFileRedis(TestLoadFromFile):
    """Test :meth:`TelescopeState.load_from_file` with redis backend."""
    def make_telescope_state(self, **kwargs) -> TelescopeState:
        return TelescopeState(RedisBackend(TabloidRedis(**kwargs)))

    def test_callback_errors_are_preserved(self) -> None:
        """Check that a redis.ConnectionError doesn't mutate into RdbParseError."""
        self.save_to_file(self.filename)
        server = fakeredis.FakeServer()
        read_ts = self.make_telescope_state(server=server)
        server.connected = False
        with self.assertRaises(redis.ConnectionError):
            read_ts.load_from_file(self.filename)



################################################################################
# Copyright (c) 2019-2020, National Research Foundation (Square Kilometre Array)
#
# Licensed under the BSD 3-Clause License (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy
# of the License at
#
#   https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import contextlib
from datetime import datetime
from typing import List, Tuple, Dict, BinaryIO, Generator, Iterable, Optional, Union, Any

import pkg_resources
import redis

from . import utils
from .backend import Backend, KeyUpdateBase, MutableKeyUpdate, ImmutableKeyUpdate, IndexedKeyUpdate
from .errors import ConnectionError, ImmutableKeyError
from .utils import KeyType, ensure_str, display_str, _PathType
try:
    from . import rdb_reader
    from .rdb_reader import BackendCallback
except ImportError as _rdb_reader_import_error:   # noqa: F841
    rdb_reader = None            # type: ignore
    BackendCallback = object     # type: ignore   # So that RedisCallback can still be defined


class RedisCallback(BackendCallback):
    """RDB callback that stores keys in :class:`redis.Redis`-like client."""

    def __init__(self, client: redis.Redis) -> None:
        super().__init__()
        self.client = client
        self._zset = {}         # type: Dict[bytes, float]
        self._hash = []         # type: List[bytes]

    def set(self, key: bytes, value: bytes, expiry: Optional[datetime], info: dict) -> None:
        self.client_busy = True
        # mypy 0.720's typeshed doesn't allow bytes keys
        self.client.set(key, value)        # type: ignore
        if expiry is not None:
            self.client.expireat(key, expiry)
        self.client_busy = False
        self.n_keys += 1

    def start_sorted_set(self, key: bytes, length: int,
                         expiry: Optional[datetime], info: dict) -> None:
        self._zset = {}
        self.n_keys += 1

    def zadd(self, key: bytes, score: float, member: bytes) -> None:
        self._zset[member] = score

    def end_sorted_set(self, key):
        self.client_busy = True
        self.client.zadd(key, self._zset)
        self.client_busy = False
        self._zset = {}

    def start_hash(self, key: bytes, length: int, expiry: Optional[datetime], info: dict) -> None:
        self._hash = []
        self.n_keys += 1

    def hset(self, key: bytes, field: bytes, value: bytes) -> None:
        self._hash.extend([field, value])

    def end_hash(self, key: bytes) -> None:
        self.client_busy = True
        # redis-py's hset doesn't support multiple key/value pairs, and its
        # hmset takes a mapping rather than interleaved keys and values. To
        # avoid the cost of building a dict and then flattening it again,
        # we execute the raw Redis command directly.
        self.client.execute_command('HSET', key, *self._hash)
        self.client_busy = False
        self._hash = []


@contextlib.contextmanager
def _handle_wrongtype() -> Generator[None, None, None]:
    """Map redis WRONGTYPE error to ImmutableKeyError.

    It also handles WRONGTYPE errors from inside scripts.
    """
    try:
        yield
    except redis.ResponseError as error:
        if (error.args[0].startswith('WRONGTYPE ')
                or (error.args[0].startswith('Error running script')
                    and 'WRONGTYPE ' in error.args[0])):
            raise ImmutableKeyError
        raise


class RedisBackend(Backend):
    """Backend for :class:`TelescopeState` using redis for storage."""

    _KEY_TYPES = {
        b'none': None,
        b'string': KeyType.IMMUTABLE,
        b'hash': KeyType.INDEXED,
        b'zset': KeyType.MUTABLE
    }

    def __init__(self, client: redis.Redis) -> None:
        self.client = client
        self._ps = self.client.pubsub(ignore_subscribe_messages=True)
        try:
            # This is the first command to the server and therefore
            # the first test of its availability
            self.client.ping()
        except (redis.TimeoutError, redis.ConnectionError) as e:
            # redis.TimeoutError: bad host
            # redis.ConnectionError: good host, bad port
            raise ConnectionError("could not connect to redis server: {}".format(e))
        self._scripts = {}     # type: Dict[str, redis.client.Script]
        for script_name in ['get', 'set_immutable', 'get_indexed', 'set_indexed',
                            'add_mutable', 'get_range']:
            script = pkg_resources.resource_string(
                'katsdptelstate', 'lua_scripts/{}.lua'.format(script_name))
            self._scripts[script_name] = self.client.register_script(script)

    def _call(self, script_name: str, *args, **kwargs) -> Any:
        return self._scripts[script_name](*args, **kwargs)

    def load_from_file(self, file: Union[_PathType, BinaryIO]) -> int:
        if rdb_reader is None:
            raise _rdb_reader_import_error   # noqa: F821
        return rdb_reader.load_from_file(RedisCallback(self.client), file)

    def __contains__(self, key: bytes) -> bool:
        # ignore due to typeshed bug: https://github.com/python/typeshed/pull/3969
        return bool(self.client.exists(key))     # type: ignore

    def keys(self, filter: bytes) -> List[bytes]:
        return self.client.keys(filter)

    def delete(self, key: bytes) -> None:
        # ignore due to typeshed bug: https://github.com/python/typeshed/pull/3969
        self.client.delete(key)                  # type: ignore

    def clear(self) -> None:
        self.client.flushdb()

    def key_type(self, key: bytes) -> Optional[KeyType]:
        type_ = self.client.type(key)
        return self._KEY_TYPES[type_]

    def set_immutable(self, key: bytes, value: bytes) -> Optional[bytes]:
        with _handle_wrongtype():
            return self._call('set_immutable', [key], [value])

    def get(self, key: bytes) -> Union[
            Tuple[None, None],
            Tuple[bytes, None],
            Tuple[bytes, float],
            Tuple[Dict[bytes, bytes], None]]:
        result = self._call('get', [key])
        if result[1] == b'none':
            return None, None
        elif result[1] == b'string':
            return result[0], None
        elif result[1] == b'zset':
            return utils.split_timestamp(result[0])
        elif result[1] == b'hash':
            it = iter(result[0])
            d = dict(zip(it, it))
            return d, None
        else:
            raise RuntimeError('Unknown key type {} for key {}'
                               .format(ensure_str(result[1]), display_str(key)))

    def add_mutable(self, key: bytes, value: bytes, timestamp: float) -> None:
        str_val = utils.pack_timestamp(timestamp) + value
        with _handle_wrongtype():
            self._call('add_mutable', [key], [str_val])

    def set_indexed(self, key: bytes, sub_key: bytes, value: bytes) -> Optional[bytes]:
        with _handle_wrongtype():
            return self._call('set_indexed', [key], [sub_key, value])

    def get_indexed(self, key: bytes, sub_key: bytes) -> Optional[bytes]:
        with _handle_wrongtype():
            result = self._call('get_indexed', [key], [sub_key])
            if result == 1:
                # Key does not exist
                raise KeyError
            else:
                return result

    def get_range(self, key: bytes, start_time: float, end_time: float,
                  include_previous: bool, include_end: bool) -> Optional[List[Tuple[bytes, float]]]:
        packed_st = utils.pack_query_timestamp(start_time, False)
        packed_et = utils.pack_query_timestamp(end_time, True, include_end)
        with _handle_wrongtype():
            ret_vals = self._call('get_range', [key], [packed_st, packed_et, int(include_previous)])
        if ret_vals is None:
            return None     # Key does not exist
        return [utils.split_timestamp(val) for val in ret_vals]

    def monitor_keys(self, keys: Iterable[bytes]) \
            -> Generator[Optional[KeyUpdateBase], Optional[float], None]:
        # Updates trigger pubsub messages to update/<key>. The format depends
        # on the key type:
        # immutable: [\x01] <value>
        # mutable:   [\x02] <timestamp> <value>
        # indexed:   \x03 <subkey-len> \n <subkey> <value>
        #            (subkey-len is ASCII text)
        # The first byte corresponds to the values in the KeyType
        # enumeration. For immutable and mutable, the leading byte is
        # currently omitted, and the key type needs to be looked up. This can
        # only be fixed once the code to handle these bytes is present in all
        # deployments.
        #
        # This can in theory cause ambiguities, because the first byte could
        # be either the key type or the start of a value or timestamp.
        # However, values are encoded and the encoding currently in use is
        # signalled with a \xFF at the start; and timestamps starting with
        # one of these bytes is infinitesimally close to zero and so will
        # not arise in practice.
        p = self.client.pubsub()
        with contextlib.closing(p):
            for key in keys:
                # ignore due to typeshed bug: https://github.com/python/typeshed/pull/3969
                p.subscribe(b'update/' + key)     # type: ignore
            timeout = yield None
            while True:
                assert timeout is not None
                message = p.get_message(timeout=timeout)
                if message is None:
                    timeout = yield None
                else:
                    assert message['channel'].startswith(b'update/')
                    key = message['channel'][7:]
                    if message['type'] == 'subscribe':
                        # An update may have occurred before our subscription
                        # happened. Also, if we lost connection and
                        # reconnected, we will get the subscribe message
                        # again.
                        timeout = yield KeyUpdateBase()
                    elif message['type'] == 'message':
                        # First check if there is a key type byte in the message.
                        data = message['data']
                        try:
                            key_type = KeyType(data[0])
                        except (ValueError, IndexError):
                            # It's the older format lacking a key type byte
                            key_type2 = self.key_type(key)
                            if key_type2 is None:
                                continue     # Key was deleted before we could retrieve the type
                            key_type = key_type2
                        else:
                            data = data[1:]  # Strip off the key type
                        if key_type == KeyType.IMMUTABLE:
                            timeout = yield ImmutableKeyUpdate(key, data)
                        elif key_type == KeyType.INDEXED:
                            # Encoding is <sub-key-length>\n<sub-key><value>
                            newline = data.find(b'\n')
                            if newline == -1:
                                raise RuntimeError('Pubsub message missing newline')
                            sub_key_len = int(data[:newline])
                            sub_key = data[newline + 1 : newline + 1 + sub_key_len]
                            value = data[newline + 1 + sub_key_len :]
                            timeout = yield IndexedKeyUpdate(key, sub_key, value)
                        elif key_type == KeyType.MUTABLE:
                            value, timestamp = utils.split_timestamp(data)
                            timeout = yield MutableKeyUpdate(key, value, timestamp)
                        else:
                            raise RuntimeError('Unhandled key type {}'.format(key_type))

    def dump(self, key: bytes) -> Optional[bytes]:
        return self.client.dump(key)
